// Matrix.h - Declarations of variables, classes, and functions for NumCPP
#ifndef MATRIX_H
#define MATRIX_H

#include <iostream>
#include <vector>
#include <limits>
#include <utility>
#include <tuple>
using namespace std;

// NumCPP namespace - to hold all the appropriate classes and functions associated with matrix operations
namespace NumCPP {

	// Constants

	/*
	Webpages consulted for constants:
	https://numpy.org/doc/stable/reference/constants.html
	https://en.cppreference.com/w/cpp/types/numeric_limits/infinity
	https://en.cppreference.com/w/cpp/types/numeric_limits/signaling_NaN
	*/
	const float g_inf = numeric_limits<float>::infinity();
	const float g_nan = numeric_limits<float>::signaling_NaN();
	const float g_ninf = -numeric_limits<float>::infinity();
	const double g_euler_constant = 2.71828182845904523536;
	const double g_pi = 3.14159265358979323846;

	// Matrix class - to be used for all matrix operations, and for storing data in a manner similar to NumPy
	template <typename T>
	class Matrix {
	private:
		// Vector of vectors to store data
		vector<vector<T>> m_data;
		// Number of rows
		int m_rows;
		// Number of columns
		int m_cols;
	public:
		// default constructor
		Matrix();
		// regular constructor
		Matrix(vector<vector<T>> a_vectors);
		// destructor
		~Matrix();
		// To set number of rows
		inline void SetRows(int a_num_rows) {
			m_rows = a_num_rows;
		}
		// To set number of columns
		inline void SetCols(int a_num_cols) {
			m_cols = a_num_cols;
		}
		// To set data
		inline void SetData(vector<vector<T>> a_vectors) {
			m_data = a_vectors;
		}
		// To access number of rows
		inline int GetRows() {
			return m_rows;
		}
		// To access number of columns
		inline int GetCols() {
			return m_cols;
		}
		// To access data
		inline vector<vector<T>> GetData() {
			return m_data;
		}
	};

	// Arithmetic operations
	template <typename T>
	Matrix<T> operator+ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> operator- (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> operator* (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> operator/ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> operator% (Matrix<T> a_matrix1, Matrix<T> a_matrix2);

	// Logical operations and overloaded << operator
	template <typename T>
	Matrix<bool> operator== (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator!= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator> (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator< (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator>= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator<= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator^ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator&& (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator|| (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> operator! (Matrix<T> a_matrix);
	template <typename T>
	ostream& operator<< (ostream& a_out, Matrix<T> a_matrix);

	// Array Creation Routines
	template <typename T>
	Matrix<T> Empty(int a_rows, int a_cols);
	template <typename T>
	Matrix<T> Empty_Like(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Eye(int a_rows, int a_columns, int a_diagonal);
	template <typename T>
	Matrix<T> Identity(int a_rows);
	template <typename T>
	Matrix<T> Ones(int a_rows, int a_cols);
	template <typename T>
	Matrix<T> Ones_Like(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Zeros(int a_rows, int a_cols);
	template <typename T>
	Matrix<T> Zeros_Like(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Full(int a_rows, int a_cols, T a_value);
	template <typename T>
	Matrix<T> Full_Like(Matrix<T> a_matrix, T a_value);
	template <typename T>
	Matrix<T> NumCppArray(vector<vector<T>> a_vectors);
	template <typename T>
	Matrix<T> LoadTxt(const string& a_filename, char a_delimiter, char a_comment);
	template <typename T>
	Matrix<T> Arange(T a_start, T a_stop, T a_step = 1);
	template <typename T>
	Matrix<T> Linspace(T a_start, T a_stop, int a_num, bool a_endpoint, bool a_retstep);
	template <typename T>
	Matrix<T> Logspace(T a_start, T a_stop, int a_num, T a_base, bool a_endpoint);
	template <typename T>
	Matrix<T> Geomspace(T a_start, T a_stop, int a_num, bool a_endpoint);
	template <typename T>
	Matrix<T> Diag(Matrix<T> a_matrix, int a_diagonal);
	template <typename T>
	Matrix<T> Tri(int a_rows, int a_cols, int a_diagonal);

	// Array Manipulation Routines
	template <typename T>
	void SaveAsCSV(Matrix<T> a_matrix);
	template <typename T>
	void CopyTo(Matrix<T>& a_destination, Matrix<T> a_source);
	template<typename T>
	vector<int> Shape(Matrix<T> a_matrix);
	template<typename T>
	vector<T> Flatten(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Transpose(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Concatenate(vector<Matrix<T>> a_matrices, int a_axis);
	template <typename T>
	Matrix<T> HStack(vector<Matrix<T>> a_matrices);
	template <typename T>
	Matrix<T> VStack(vector<Matrix<T>> a_matrices);
	template <typename T>
	Matrix<T> Column_Stack(vector<Matrix<T>> a_matrices);
	template <typename T>
	Matrix<T> Row_Stack(vector<Matrix<T>> a_matrices);
	template <typename T>
	vector<T> Unique(Matrix<T> a_matrix, bool a_unique_indices, bool a_unique_counts);
	template <typename T>
	Matrix<T> Flip(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Repeat(Matrix<T> a_matrix, int a_repeats, int a_axis);
	template <typename T>
	Matrix<T> Trim_Zeros(Matrix<T> a_matrix, string a_trim);
	template <typename T>
	void Delete(Matrix<T>& a_matrix, int a_index, int a_axis);
	template <typename T>
	void Insert(Matrix<T>& a_matrix, int a_index, T a_value, int a_axis);
	template <typename T>
	void Append(Matrix<T>& a_matrix_one, Matrix<T> a_matrix_two, int a_axis);

	// Logic functions
	template <typename T>
	Matrix<bool> Logical_And(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Logical_Or(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Logical_XOR(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Logical_Not(Matrix<T> a_matrix);
	template <typename T>
	Matrix<bool> Greater(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Greater_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Less(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Less_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Not_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<bool> Is_Finite(Matrix<T> a_matrix);
	template <typename T>
	Matrix<bool> Is_Inf(Matrix<T> a_matrix);
	template <typename T>
	Matrix<bool> Is_Nan(Matrix<T> a_matrix);
	template <typename T>
	Matrix<bool> Contains(Matrix<T> a_matrix, T a_value);
	template <typename T>
	Matrix<bool> Is_NegInf(Matrix<T> a_matrix);
	template <typename T>
	Matrix<bool> Is_PosInf(Matrix<T> a_matrix);

	// Mathematical functions - Sum and Product
	template <typename T>
	vector<T> Sum(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	vector<T> Prod(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	vector<T> NanSum(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	vector<T> NanProd(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> CumSum(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> CumProd(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> NanCumSum(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> NanCumProd(Matrix<T> a_matrix, int a_axis);

	// Mathematical functions - Exponents and logarithms
	template <typename T>
	Matrix<T> Power(Matrix<T> a_matrix, T a_base);
	template <typename T>
	Matrix<T> Exp(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Exp2(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> ExpM1(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Log(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Log10(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Log2(Matrix<T> a_matrix);

	// Mathematical functions - Rational routines
	template <typename T>
	int LCM(Matrix<T> a_matrix);
	template <typename T>
	int GCD(Matrix<T> a_matrix);

	// Helper functions
	inline int Find_GCD(int a_num1, int a_num2);
	inline int Factorial(int a_num);
	inline int Binomial_Coefficient(int a_num1, int a_num2);

	// Mathematical functions - Trigonometric routines
	template <typename T>
	Matrix<T> Sin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
	template <typename T>
	Matrix<T> Cos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
	template <typename T>
	Matrix<T> Tan(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcSin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcCos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcTan(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> Sinh(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> Cosh(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> Tanh(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcSinh(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcCosh(Matrix<T> a_matrix, int a_decimal_places);
	template <typename T>
	Matrix<T> ArcTanh(Matrix<T> a_matrix, int a_decimal_places);

	// Mathematical functions - Linear algebra
	template <typename T>
	Matrix<T> Inner_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> Outer_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> Mat_Mul(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	Matrix<T> Dot(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
	template <typename T>
	pair<T, bool> Eigenvalue(Matrix<T> a_matrix, Matrix<T> a_eigenvector);
	template <typename T>
	void Swap(vector<vector<T>>& a_vector, int a_row_index_one, int a_column_index_one, int a_row_index_two, int a_column_index_two);
	template <typename T>
	T Determinant(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Rref(Matrix<T> a_matrix);
	template <typename T>
	Matrix<T> Matrix_Power(Matrix<T> a_matrix, int a_power);
	template <typename T>
	Matrix<T> Inverse(Matrix<T> a_matrix);

	// Mathematical functions - Statistics
	template <typename T>
	Matrix<T> Median(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Mean(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Average(Matrix<T> a_matrix, vector<T> a_weights, int a_axis);
	template <typename T>
	Matrix<T> Std(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Var(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> PTP(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Percentile(Matrix<T> a_matrix, double a_percentile, int a_axis);

	// Sorting, searching, and counting
	template <typename T>
	int Partition(vector<T>& a_vector, int a_start_index, int a_end_index);
	template <typename T>
	void Quick_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);
	template <typename T>
	void Heapify(vector<T>& a_vector, int a_vector_size, int a_index);
	template <typename T>
	void Heap_Sort(vector<T>& a_vector);
	template <typename T>
	void Merge(vector<T>& a_vector, int a_start_index, int a_middle_index, int a_end_index);
	template <typename T>
	void Merge_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);
	template <typename T>
	void Insertion_Sort(vector<T>& a_vector, int a_left_index, int a_right_index);
	template <typename T>
	void Tim_Sort(vector<T>& a_vector, int a_run_size);
	template <typename T>
	Matrix<T> Sort(Matrix<T> a_matrix, string a_kind, int a_axis);
	template <typename T>
	Matrix<int> ArgMax(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<int> ArgMin(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<int> NanArgMax(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<int> NanArgMin(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	tuple<vector<int>, vector<int>> NonZero(Matrix<T> a_matrix);
	template <typename T>
	Matrix<int> FlatNonZero(Matrix<T> a_matrix);
	template <typename T>
	Matrix<int> CountNonZero(Matrix<T> a_matrix, int a_axis);
	template <typename T>
	Matrix<T> Where(Matrix<bool> a_condition_matrix, Matrix<T> a_true, Matrix<T> a_false);
}
#endif