// Matrix.cpp - Implementations of functions for NumCPP
#ifndef MATRIX_CPP
#define MATRIX_CPP

#include "Matrix.h"
#include <type_traits>
#include <cstdlib>
#include <fstream>
#include <sstream>
#include <cmath>
#include <algorithm>

namespace NumCPP {
	/**/
	/*
	NAME

		Matrix - Default constructor for the Matrix class.

	SYNOPSIS

		Matrix<T>::Matrix();

	DESCRIPTION

		This function is the default constructor for the Matrix class.
		The implementation was thought of by me, but GitHub Copilot
		was used to ensure that no syntax errors were made (i.e., missing semicolons at the end of
		each statement). Same goes for almost all of the
		other functions in this file unless otherwise noted.

	RETURNS

		NONE
	*/
	/**/
	template <typename T>
	Matrix<T>::Matrix() {
		m_rows = 0;
		m_cols = 0;
		m_data = vector<vector<T>>();
	}
	/*Matrix<T>::Matrix();*/

	/**/
	/*
	NAME

		Matrix - Regular constructor for the Matrix class.

	SYNOPSIS

		Matrix<T>::Matrix(vector<vector<T>> a_vectors);
			a_vectors --> The vector of vectors containing the data that will be stored in the Matrix.

	DESCRIPTION

		This function is the regular constructor for the Matrix class.

	RETURNS

		NONE
	*/
	/**/
	template <typename T>
	Matrix<T>::Matrix(vector<vector<T>> a_vectors) {
		// Make sure all rows have the same number of columns
		bool valid_dimensions = true;
		for (int row_index = 0; row_index < a_vectors.size(); row_index++) {
			if (row_index > 0) {
				if (a_vectors[row_index].size() != a_vectors[row_index - 1].size()) {
					valid_dimensions = false;
					break;
				}
			}
			else {
				continue;
			}
		}
		if (valid_dimensions == true) {
			// If we get here, then all rows have the same number of columns
			m_rows = a_vectors.size();
			m_cols = a_vectors[0].size();
			m_data = a_vectors;
		}
		else {
			// If we get here, then not all rows have the same number of columns
			cout << "Error: Invalid dimensions for matrix." << endl;
			cout << "Setting matrix to Empty." << endl;
			m_rows = 0;
			m_cols = 0;
			m_data = vector<vector<T>>();
		}
	}
	/*Matrix<T>::Matrix(vector<vector<T>> a_vectors);*/

	/**/
	/*
	NAME

		~Matrix - Destructor for the Matrix class.

	SYNOPSIS

		Matrix<T>::~Matrix();

	DESCRIPTION

		This function is the destructor for the Matrix class.

	RETURNS

		NONE
	*/
	/**/
	template <typename T>
	Matrix<T>::~Matrix() {
	}
	/*Matrix<T>::~Matrix();*/

	/**/
	/*
	NAME

		operator+ - Add together two Matrix instances.

	SYNOPSIS

		Matrix<T> operator+ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function adds together the data stored in the two Matrix instances.
		To help get an understanding of how to use the 'is_same' function, I consulted
		the following website: https://www.geeksforgeeks.org/stdis_same-template-in-c-with-examples/
		Same with the rest of the overloaded operators that use the 'is_same' function.

	RETURNS

		A new instance of the Matrix class with its data set equal to the element-wise sums
		of the two initial Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> operator+ (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not trying to add matrices of boolean values
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform arithmetic on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					row.push_back(a_matrix1.GetData()[row_index][column_index] + a_matrix2.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> operator+ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/


	/**/
	/*
	NAME

		operator- - Subtract two Matrix instances.

	SYNOPSIS

		Matrix<T> operator- (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function subtracts the data stored in the two Matrix instances.

	RETURNS

		A new instance of the Matrix class with its data set equal to the element-wise differences
		of the two initial Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> operator- (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not trying to subtract matrices of boolean values
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform arithmetic on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// Make sure the dimensions are equal
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					row.push_back(a_matrix1.GetData()[row_index][column_index] - a_matrix2.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> operator- (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator* - Multiply two Matrix instances.

	SYNOPSIS

		Matrix<T> operator* (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function multiplies the data stored in the two Matrix instances.

	RETURNS

		A new instance of the Matrix class with its data set equal to the element-wise products
		of the two initial Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> operator* (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not trying to multiply matrices of boolean values
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform arithmetic on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// Make sure the dimensions are equal
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					row.push_back(a_matrix1.GetData()[row_index][column_index] * a_matrix2.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> operator* (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator/ - Divide two Matrix instances.

	SYNOPSIS

		Matrix<T> operator/ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function divides the data stored in the two Matrix instances.

	RETURNS

		A new instance of the Matrix class with its data set equal to the element-wise quotients
		of the two initial Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> operator/ (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not trying to divide matrices of boolean values
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform arithmetic on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// Make sure the dimensions are equal
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					if (a_matrix2.GetData()[row_index][column_index] == 0) {
						cout << "Error: Cannot divide by zero." << endl;
						cout << "Returning Empty matrix." << endl;
						return Matrix<T>();
					}
					row.push_back(a_matrix1.GetData()[row_index][column_index] / a_matrix2.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> operator/ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator% - Obtain the remainders between two Matrix instances.

	SYNOPSIS

		Matrix<T> operator% (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function computes the remainders of the data stored in the two Matrix instances.

	RETURNS

		A new instance of the Matrix class with its data set equal to the element-wise remainders
		of the two initial Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> operator% (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not trying to compute the remainders of boolean matrices
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform arithmetic on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// Make sure the dimensions are equal
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					if (a_matrix2.GetData()[row_index][column_index] == 0) {
						cout << "Error: Cannot perform % operation with 0." << endl;
						cout << "Returning Empty matrix." << endl;
						return Matrix<T>();
					}
					row.push_back(a_matrix1.GetData()[row_index][column_index] % a_matrix2.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> operator% (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator== - Determine if elements in two Matrix instances are the same.

	SYNOPSIS

		Matrix<bool> operator== (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function checks for equality between the elements being stored at the same indices in
		both Matrix instances.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the elements at each corresponding index are the same.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator== (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] == a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator== (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator!= - Determine if elements in two Matrix instances are not the same.

	SYNOPSIS

		Matrix<bool> operator!= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function compares the data stored in the two Matrix instances to see if they are not the same.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the elements located at the same index in both Matrix instances are different.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator!= (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] != a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator!= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator> - Determine if elements in one Matrix instance are greater than those in another Matrix instance.

	SYNOPSIS

		Matrix<bool> operator> (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function compares two Matrix instances to see if the elements in the first instance are greater
		than those in the second instance.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the element located at one index in the first Matrix instance is greater than one located at the same index
		in the second Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator> (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with boolean matrices
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform > operation on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] > a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator> (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator< - Determine if elements in one Matrix instance are less than those in another Matrix instance.

	SYNOPSIS

		Matrix<bool> operator< (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function compares two Matrix instances to see if the elements in the first instance are less
		than those in the second instance.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the element located at one index in the first Matrix instance is less than one located at the same index
		in the second Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator< (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with boolean matrices
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform < operation on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] < a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator< (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator>= - Determine if elements in one Matrix instance are greater than or equal to those in another Matrix instance.

	SYNOPSIS

		Matrix<bool> operator>= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function compares two Matrix instances to see if the elements in the first instance are greater
		than or equal to those in the second instance.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the element located at one index in the first Matrix instance is greater than or equal to one located at the same index
		in the second Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator>= (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with boolean matrices
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform >= operation on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] >= a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator>= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator<= - Determine if elements in one Matrix instance are less than or equal to those in another Matrix instance.

	SYNOPSIS

		Matrix<bool> operator<= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function compares two Matrix instances to see if the elements in the first instance are less
		than or equal to those in the second instance.

	RETURNS

		A new instance of the Matrix class with its data set equal to numbers representing whether or not
		the element located at one index in the first Matrix instance is less than or equal to one located at the same index
		in the second Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator<= (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with boolean matrices
		if (is_same<T, bool>::value == true) {
			cout << "Error: Cannot perform <= operation on boolean matrices." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions are the same
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					// 1 represents true, 0 represents false
					if (a_matrix1.GetData()[row_index][column_index] <= a_matrix2.GetData()[row_index][column_index]) {
						row.push_back(1);
					}
					else {
						row.push_back(0);
					}
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator<= (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator^ - Perform the XOR operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> operator^ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function performs the XOR operation on the data stored in the two Matrix instances.

	RETURNS

		A Matrix instance containing the results of the XOR operation on the data stored in the two Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator^ (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with matrices of floating point numbers
		if (is_same<T, double>::value == true || is_same<T, float>::value == true) {
			cout << "Error: Cannot perform ^ operation on matrices of floating point numbers." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions of both matrices are the same before progressing
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					bool result = a_matrix1.GetData()[row_index][column_index] ^ a_matrix2.GetData()[row_index][column_index];
					row.push_back(result);
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator^ (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator&& - Perform the logical AND operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> operator&& (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function performs the AND operation on the data stored in the two Matrix instances.

	RETURNS

		A Matrix instance containing the results of the AND operation on the data stored in the two Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator&& (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with matrices of floating point numbers
		if (is_same<T, double>::value == true || is_same<T, float>::value == true) {
			cout << "Error: Cannot perform && operation on matrices of floating point numbers." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions of both matrices are the same before progressing
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					bool result = a_matrix1.GetData()[row_index][column_index] && a_matrix2.GetData()[row_index][column_index];
					row.push_back(result);
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator&& (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator|| - Perform the logical OR operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> operator|| (Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance being considered
			a_matrix2 -> The second Matrix instance being considered

	DESCRIPTION

		This function performs the OR operation on the data stored in the two Matrix instances.

	RETURNS

		A Matrix instance containing the results of the OR operation on the data stored in the two Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator|| (Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// Make sure we are not working with matrices of floating point numbers
		if (is_same<T, double>::value == true || is_same<T, float>::value == true) {
			cout << "Error: Cannot perform || operation on matrices of floating point numbers." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		// Make sure the dimensions of both matrices are the same before progressing
		if (a_matrix1.GetRows() == a_matrix2.GetRows() && a_matrix1.GetCols() == a_matrix2.GetCols()) {
			vector<vector<bool>> result;
			for (int row_index = 0; row_index < a_matrix1.GetRows(); row_index++) {
				vector<bool> row;
				for (int column_index = 0; column_index < a_matrix1.GetCols(); column_index++) {
					bool result = a_matrix1.GetData()[row_index][column_index] || a_matrix2.GetData()[row_index][column_index];
					row.push_back(result);
				}
				result.push_back(row);
			}
			return Matrix<bool>(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}
	}
	/*Matrix<bool> operator|| (Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		operator! - Perform the logical NOT operation on a Matrix instance.

	SYNOPSIS

		Matrix<bool> operator! (Matrix<T> a_matrix);
			a_matrix -> The Matrix instance being considered

	DESCRIPTION

		This function performs the NOT operation on the data stored in the Matrix instance.

	RETURNS

		A Matrix instance containing the result of the NOT operation on the data stored in the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<bool> operator! (Matrix<T> a_matrix) {
		// Make sure we are not working with a matrix of strings
		if (is_same<T, string>::value == true) {
			cout << "Error: Cannot perform ! operation on matrices of strings." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<bool>();
		}

		vector<vector<bool>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<bool> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				bool current_value = !(a_matrix.GetData()[current_row][current_column]);
				row.push_back(current_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<bool> operator! (Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		operator<< - Store the data stored in a Matrix instance in an 'ostream' object

	SYNOPSIS

		ostream& operator<< (ostream& a_out, Matrix<T> a_matrix);
			a_out -> A reference to an 'ostream' object where the Matrix instance's data will be stored
			a_matrix -> The Matrix instance being considered

	DESCRIPTION

		This function compares the data stored in a Matrix instance in an 'ostream' object
		so it can be printed out when we call on the instance like this:
		cout << matrix_instance << endl;

	RETURNS

		A reference to an 'ostream' object where the Matrix instance's data has been stored.
	*/
	/**/
	template <typename T>
	ostream& operator<< (ostream& a_out, Matrix<T> a_matrix) {
		// Determine what to store in the ostream object depending on the dimensions of the matrix
		if (a_matrix.GetRows() == 0 && a_matrix.GetCols() == 0) {
			a_out << "{}";
		}
		else {
			for (int row_index = 0; row_index < a_matrix.GetRows(); row_index++) {
				for (int column_index = 0; column_index < a_matrix.GetCols(); column_index++) {
					a_out << a_matrix.GetData()[row_index][column_index] << " ";
				}
				a_out << endl;
			}
		}
		return a_out;
	}
	/*ostream& operator<< (ostream& a_out, Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Empty - Initialize a Matrix instance without initializing its data.

	SYNOPSIS

		Matrix<T> Empty(int a_rows, int a_cols);
			a_rows -> The number of rows in the Matrix instance
			a_cols -> The number of columns in the Matrix instance

	DESCRIPTION

		This function initializes a Matrix instance without initializing its data.
		To accomplish this, we use either the 'rand()' function to generate random numbers,
		or boolean values, and store them in the Matrix instance.

	RETURNS

		A Matrix instance with its data uninitialized.
	*/
	/**/
	template <typename T>
	Matrix<T> Empty(int a_rows, int a_cols) {
		if (a_rows < 0 || a_cols < 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		if (is_same<T, bool>::value == true) {
			// If we get here, then we are working with a boolean matrix.
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_rows; row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_cols; column_index++) {
					int boolean = rand() % 2;
					T value;
					if (boolean == 0) {
						value = false;
					}
					else {
						value = true;
					}
					row.push_back(value);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are working with a matrix of numbers.
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_rows; row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_cols; column_index++) {
					T random_number = (T)rand();
					row.push_back(random_number);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Empty(int a_rows, int a_cols);*/

	/**/
	/*
	NAME

		Empty_Like - Create an uninitialized Matrix instance with the same dimensions as another Matrix instance.

	SYNOPSIS

		Matrix<T> Empty_Like(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose dimensions will be used to create the new Matrix instance

	DESCRIPTION

		This function calls the 'Empty()' function to create a Matrix instance with the same dimensions
		as the Matrix instance passed in as a parameter.

	RETURNS

		A Matrix instance with the same dimensions as the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Empty_Like(Matrix<T> a_matrix) {
		Matrix<T> result = Empty<T>(a_matrix.GetRows(), a_matrix.GetCols());
		return result;
	}
	/*Matrix<T> Empty_Like(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Eye - Create a Matrix instance, with a diagonal of 1s, and 0s everywhere else.

	SYNOPSIS

		Matrix<T> Eye(int a_rows, int a_columns, int a_diagonal);
			a_rows -> The number of rows in the Matrix instance
			a_columns -> The number of columns in the Matrix instance
			a_diagonal -> The diagonal of 1s in the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance, with 1s on the specified diagonal, and 0s everywhere else.

	RETURNS

		A Matrix instance with 1s on the specified diagonal, and 0s everywhere else.
	*/
	/**/
	template <typename T>
	Matrix<T> Eye(int a_rows, int a_columns, int a_diagonal) {
		vector<vector<T>> result;
		int total_columns;
		if (a_rows <= 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		// If the user passes in a negative number or 0 for the number of columns, then we set the number of columns to the number of rows.
		if (a_columns <= 0) {
			total_columns = a_rows;
		}
		else {
			total_columns = a_columns;
		}

		for (int row_index = 0; row_index < a_rows; row_index++) {
			vector<T> row;
			for (int column_index = 0; column_index < total_columns; column_index++) {
				if (column_index == row_index + a_diagonal) {
					row.push_back(1);
				}
				else {
					row.push_back(0);
				}
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Eye(int a_rows, int a_columns, int a_diagonal);*/

	/**/
	/*
	NAME

		Identity - Create the identity matrix.

	SYNOPSIS

		Matrix<T> Identity(int a_rows);
			a_rows -> The number of rows in the Matrix instance

	DESCRIPTION

		This function creates a square Matrix instance with 1s on the diagonal, and 0s everywhere else.
		To accomplish this, we call the 'Eye()' function by passing in the number of rows as the number of columns,
		and 0 as the diagonal.

	RETURNS

		A Matrix instance containing the identity matrix, based on the number of rows passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Identity(int a_rows) {
		return Eye<T>(a_rows, a_rows, 0);
	}
	/*Matrix<T> Identity(int a_rows);*/

	/**/
	/*
	NAME

		Ones - Create a matrix of 1s.

	SYNOPSIS

		Matrix<T> Ones(int a_rows, int a_columns);
			a_rows -> The number of rows in the Matrix instance
			a_columns -> The number of columns in the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with 1s everywhere.

	RETURNS

		A Matrix instance where every element is equal to 1.
	*/
	/**/
	template <typename T>
	Matrix<T> Ones(int a_rows, int a_columns) {
		vector<vector<T>> result;
		if (a_rows < 0 || a_columns < 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else {
			for (int row_index = 0; row_index < a_rows; row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_columns; column_index++) {
					row.push_back(1);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Ones(int a_rows, int a_columns);*/

	/**/
	/*
	NAME

		Ones_Like - Create a matrix of 1s with the same dimensions as another Matrix instance.

	SYNOPSIS

		Matrix<T> Ones_Like(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose dimensions will be used to create the new Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with 1s everywhere, with the same dimensions as the Matrix instance passed in as a parameter.

	RETURNS

		A Matrix instance where every element is equal to 1, and has the same dimensions as the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Ones_Like(Matrix<T> a_matrix) {
		return Ones<T>(a_matrix.GetRows(), a_matrix.GetCols());
	}
	/*Matrix<T> Ones_Like(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Zeros - Create a matrix of 0s.

	SYNOPSIS

		Matrix<T> Zeros(int a_rows, int a_columns);
			a_rows -> The number of rows in the Matrix instance
			a_columns -> The number of columns in the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with 0s everywhere.

	RETURNS

		A Matrix instance where every element is equal to 0.
	*/
	/**/
	template <typename T>
	Matrix<T> Zeros(int a_rows, int a_columns) {
		if (a_rows < 0 || a_columns < 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_rows; row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_columns; column_index++) {
					row.push_back(0);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Zeros(int a_rows, int a_columns);*/

	/**/
	/*
	NAME

		Zeros_Like - Create a matrix of 0s with the same dimensions as another Matrix instance.

	SYNOPSIS

		Matrix<T> Zeros_Like(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose dimensions will be used to create the new Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with 0s everywhere, with the same dimensions as the Matrix instance passed in as a parameter.

	RETURNS

		A Matrix instance where every element is equal to 0, and has the same dimensions as the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Zeros_Like(Matrix<T> a_matrix) {
		return Zeros<T>(a_matrix.GetRows(), a_matrix.GetCols());
	}
	/*Matrix<T> Zeros_Like(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Full - Create a matrix filled with a specified value.

	SYNOPSIS

		Matrix<T> Full(int a_rows, int a_columns, T a_value);
			a_rows -> The number of rows in the Matrix instance
			a_columns -> The number of columns in the Matrix instance
			a_value -> The value to be stored in every index of the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with every element equal to a specified value.

	RETURNS

		A Matrix instance where every element is equal to a specified value.
	*/
	/**/
	template <typename T>
	Matrix<T> Full(int a_rows, int a_columns, T a_value) {
		if (a_rows < 0 || a_columns < 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning empty matrix." << endl;
			return Matrix<T>();
		}
		else {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_rows; row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_columns; column_index++) {
					row.push_back(a_value);
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Full(int a_rows, int a_columns, T a_value);*/

	/**/
	/*
	NAME

		Full_Like - Create a matrix filled with a specified value, with the same dimensions as another Matrix instance.

	SYNOPSIS

		Matrix<T> Full_Like(Matrix<T> a_matrix, T a_value);
			a_matrix -> The Matrix instance whose dimensions will be used to create the new Matrix instance
			a_value -> The value to be stored in every index of the new Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with every element equal to a specified value, with the same dimensions as the Matrix instance passed in as a parameter.

	RETURNS

		A Matrix instance where every element is equal to a specified value, and has the same dimensions as the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Full_Like(Matrix<T> a_matrix, T a_value) {
		return Full<T>(a_matrix.GetRows(), a_matrix.GetCols(), a_value);
	}
	/*Matrix<T> Full_Like(Matrix<T> a_matrix, T a_value);*/

	/**/
	/*
	NAME

		NumCppArray - Create a Matrix instance from a vector of vectors.

	SYNOPSIS

		Matrix<T> NumCppArray(vector<vector<T>> a_vectors);
			a_vectors -> The vector of vectors containing the data to be stored in the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance from a vector of vectors.

	RETURNS

		A Matrix instance containing the data stored in the vector of vectors passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> NumCppArray(vector<vector<T>> a_vectors) {
		return Matrix<T>(a_vectors);
	}
	/*Matrix<T> NumCppArray(vector<vector<T>> a_vectors);*/

	/**/
	/*
	NAME

		LoadTxt - Create a Matrix instance from a text file.

	SYNOPSIS

		Matrix<T> LoadTxt(const string& a_filename, char a_delimiter, char a_comment);
			a_filename -> The name of the file containing the data to be stored in the Matrix instance
			a_delimiter -> The character that separates the data in the file
			a_comment -> The character that indicates a comment in the file

	DESCRIPTION

		This function creates a Matrix instance by reading in data from a text file.
		Most of the code in this function was borrowed from a conversation with ChatGPT
		where I asked it how to read in data from a text file and store it in a vector.
		Everything else that involved adding those vectors to another vector, and then
		creating a Matrix instance from that vector was written by me.

	RETURNS

		A Matrix instance containing the data stored in the text file passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> LoadTxt(const string& a_filename, char a_delimiter, char a_comment) {
		vector<vector<T>> result;
		ifstream file(a_filename);
	
		// Make sure the file exists
		if (file.is_open()) {
			string line;
			while (getline(file, line)) {
				istringstream ss(line);
				string token;
				vector<T> row;
				while (getline(ss, token, a_delimiter)) {
					if (line[0] == a_comment) {
						// If we get here, then we are dealing with a comment, so we skip it
						continue;
					}
					istringstream converter(token);
					T value;
					converter >> value;
					row.push_back(value);
				}
				// Make sure we are not adding an Empty row to the vector, since that would mess up the dimensions of the Matrix instance
				if (row.size() > 0) {
					result.push_back(row);
				}
			}
			file.close();
			return Matrix<T>(result);
		}
		else {
			cout << "Unable to open file: " << a_filename << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
	}
	/*Matrix<T> LoadTxt(const string& a_filename, char a_delimiter, char a_comment);*/

	/**/
	/*
	NAME

		Arange - Create a Matrix instance with evenly spaced values within a given interval.

	SYNOPSIS

		Matrix<T> Arange(T a_start, T a_stop, T a_step);
			a_start -> The starting value of the interval
			a_stop -> The ending value of the interval
			a_step -> The step value of the interval

	DESCRIPTION

		This function creates a Matrix instance with evenly spaced values within a given interval.
		But before we do this, we check for the following cases:
			1. If the starting value is greater than the ending value, and the step value is positive, then we return an Empty matrix.
			2. If the starting value is less than the ending value, and the step value is negative, then we return an Empty matrix.
			3. If the step value is equal to zero, then we return an Empty matrix.

	RETURNS

		A Matrix instance containing evenly spaced values within a given interval.
	*/
	/**/
	template <typename T>
	Matrix<T> Arange(T a_start, T a_stop, T a_step) {
		// Check for case #1 described in the documentation
		if ((a_start > a_stop) && (a_step > 0)) {
			return Matrix<T>();
		}

		// Check for case #2 described in the documentation
		else if ((a_start < a_stop) && (a_step < 0)) {
			return Matrix<T>();
		}

		// Check for case #3 described in the documentation
		else if (a_step == 0) {
			cout << "Step cannot be equal to zero." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		// If we get here, then we are dealing with valid interval and step values
		else {
			vector<vector<T>> result;
			vector<T> row;
			T temp = a_start;
			while (temp < a_stop) {
				row.push_back(temp);
				temp += a_step;
			}
			result.push_back(row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Arange(T a_start, T a_stop, T a_step);*/

	/**/
	/*
	NAME

		Linspace - Returns evenly spaced numbers over a specified interval.

	SYNOPSIS

		Matrix<T> Linspace(T a_start, T a_stop, int a_num, bool a_endpoint, bool a_retstep);
			a_start -> The starting value of the interval
			a_stop -> The ending value of the interval
			a_num -> The number of samples to generate
			a_endpoint -> Whether or not to include the endpoint in the interval
			a_retstep -> Whether or not to return the step value

	DESCRIPTION

		This function behaves similar to the 'Arange' function, but the main difference is that
		in 'Arange', we determine the step value, whereas in 'Linspace', we determine the number of samples.
		The code was written with the help of the following link: https://gist.github.com/mortenpi/f20a93c8ed3ee7785e65

	RETURNS

		A Matrix instance containing evenly spaced values within a given interval, based on how many samples we want.
	*/
	/**/
	template <typename T>
	Matrix<T> Linspace(T a_start, T a_stop, int a_num, bool a_endpoint, bool a_retstep) {
		T step;

		// Check the value of 'a_num' before we proceed
		if (a_num < 0) {
			cout << "Error: Number of samples must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else if (a_num == 0) {
			return Matrix<T>();
		}
		else if (a_num == 1) {
			vector<vector<T>> result;
			vector<T> row;
			row.push_back(a_start);
			result.push_back(row);
			if (a_retstep == true) {
				cout << "Step taken in Linspace: " << 0 << endl;
			}
			return Matrix<T>(result);
		}

		// Check the value of 'a_endpoint' to determine the step value
		if (a_endpoint == true) {
			step = (a_stop - a_start) / (a_num - 1);
		}
		else {
			step = (a_stop - a_start) / a_num;
		}

		// Create our Matrix instance
		vector<vector<T>> result;
		vector<T> row;
		T temp = a_start;
		for (int current_index = 0; current_index < a_num; current_index++) {
			row.push_back(temp);
			temp += step;
		}
		result.push_back(row);
		if (a_retstep == true) {
			cout << "Step taken in Linspace: " << step << endl;
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Linspace(T a_start, T a_stop, int a_num, bool a_endpoint, bool a_retstep);*/

	/**/
	/*
	NAME

		Logspace - Returns evenly spaced numbers over a specified interval on a log scale.

	SYNOPSIS

		Matrix<T> Logspace(T a_start, T a_stop, int a_num, T a_base, bool a_endpoint);
			a_start -> The starting value of the interval
			a_stop -> The ending value of the interval
			a_num -> The number of samples to generate
			a_base -> The base of the log space
			a_endpoint -> Whether or not to include the endpoint in the interval

	DESCRIPTION

		This function behaves similar to the 'Linspace' function, but the main difference is that
		in 'Logspace', the numbers are evenly spaced on a log scale, whereas in 'Linspace', the numbers
		are evenly spaced on a linear scale.
		The code was written with the help of the following link: https://github.com/compute-io/Logspace/blob/master/lib/index.js

	RETURNS

		A Matrix instance containing evenly spaced values within a given interval on a log scale, based on how many samples we want.
	*/
	/**/
	template <typename T>
	Matrix<T> Logspace(T a_start, T a_stop, int a_num, T a_base, bool a_endpoint) {
		T power;

		// Check the value of 'a_num' before we proceed
		if (a_num < 0) {
			cout << "Error: Number of samples must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else if (a_num == 0) {
			return Matrix<T>();
		}
		else if (a_num == 1) {
			vector<vector<T>> result;
			vector<T> row;
			row.push_back(pow(a_base, a_start));
			result.push_back(row);
			return Matrix<T>(result);
		}

		// Check the value of 'a_endpoint' to determine the value of 'power'
		if (a_endpoint == true) {
			power = (a_stop - a_start) / (a_num - 1);
		}
		else {
			power = (a_stop - a_start) / a_num;
		}

		// Create our Matrix instance
		vector<vector<T>> result;
		vector<T> row;
		T temp = a_start;
		for (int current_index = 0; current_index < a_num; current_index++) {
			row.push_back(pow(a_base, temp));
			temp += power;
		}
		result.push_back(row);
		return Matrix<T>(result);
	}
	/*Matrix<T> Logspace(T a_start, T a_stop, int a_num, T a_base, bool a_endpoint);*/

	/**/
	/*
	NAME

		Geomspace - Returns evenly spaced numbers over a specified interval based on a geometric sequence.

	SYNOPSIS

		Matrix<T> Geomspace(T a_start, T a_stop, int a_num, bool a_endpoint);
			a_start -> The starting value of the interval
			a_stop -> The ending value of the interval
			a_num -> The number of samples to generate
			a_endpoint -> Whether or not to include the endpoint in the interval

	DESCRIPTION

		This function behaves similar to the 'Logspace' function, but the main difference is that
		in 'Geomspace', the numbers are evenly spaced based on a geometric sequence, whereas in 'Logspace',
		the numbers are evenly spaced on a log scale.
		To determine what the common ratio is based on the start and stop values, the following link was consulted:
		https://www.geeksforgeeks.org/how-to-find-common-ratio-with-first-and-last-terms/
		And to help with figuring out how to use the 'round()' function in the context of 'Geomspace', the following link was consulted:
		https://www.geeksforgeeks.org/round-in-cpp/

	RETURNS

		A Matrix instance containing evenly spaced values within a given interval based on a geometric sequence, based on how many samples we want.
	*/
	/**/
	template <typename T>
	Matrix<T> Geomspace(T a_start, T a_stop, int a_num, bool a_endpoint) {
		// Check the value of 'a_num' before we proceed
		if (a_num == 0) {
			return Matrix<T>();
		}
		else if (a_num < 0) {
			cout << "Error: Number of samples must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else if (a_num == 1) {
			vector<vector<T>> result;
			vector<T> row;
			row.push_back(a_start);
			result.push_back(row);
			return Matrix<T>(result);
		}

		// Make sure the starting and stopping values are not less than or equal to 0
		if ((a_start <= 0) || (a_stop <= 0)) {
			cout << "Error: Geometric sequence cannot include numbers less than or equal to zero." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else if (a_start == a_stop) {
			vector<vector<T>> result;
			vector<T> row;
			for (int current_index = 0; current_index < a_num; current_index++) {
				row.push_back(a_start);
			}
			result.push_back(row);
			return Matrix<T>(result);
		}

		// Determine the value of 'power_exponent' based on the value of 'a_endpoint'
		double power_exponent;
		if (a_endpoint == true) {
			power_exponent = 1.0 / (a_num - 1);
		}
		else {
			power_exponent = 1.0 / a_num;
		}

		// Determine the value of 'power' based on whether or not the starting value is less than the stopping value
		double power;
		if (a_start < a_stop) {
			power = pow(a_stop / a_start, power_exponent);
		}
		else {
			power = pow(a_start / a_stop, power_exponent);
		}

		// Create our Matrix instance
		vector<vector<T>> result;
		if (a_start < a_stop) {
			vector<T> row;
			for (int current_index = 0; current_index < a_num; current_index++) {
				row.push_back(round(a_start * pow(power, current_index)));
			}
			result.push_back(row);
		}
		else {
			vector<T> row;
			for (int current_index = 0; current_index < a_num; current_index++) {
				row.push_back(round(a_start / pow(power, current_index)));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Geomspace(T a_start, T a_stop, int a_num, bool a_endpoint);*/

	/**/
	/*
	NAME

		Diag - Extract a diagonal from a Matrix instance.

	SYNOPSIS

		Matrix<T> Diag(Matrix<T> a_matrix, int a_diagonal);
			a_matrix -> The Matrix instance being considered
			a_diagonal -> The diagonal of the Matrix instance to be extracted

	DESCRIPTION

		This function gets the diagonal of a Matrix instance based on the diagonal number passed in as a parameter.
		To help with figuring out how to check if we are on the diagonal, a conversation with Microsoft Copilot was consulted.

	RETURNS

		A Matrix instance containing numbers located on the specified diagonal of the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Diag(Matrix<T> a_matrix, int a_diagonal) {
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				if (current_column == current_row + a_diagonal) {
					// If we get here, then we are on the diagonal
					row.push_back(a_matrix.GetData()[current_row][current_column]);
				}
				else {
					continue;
				}
			}
			if (row.size() > 0) {
				result.push_back(row);
			}
		}

		// If the result vector is empty, then we return an Empty matrix. Otherwise, we return the matrix containing the diagonal elements.
		if (result.size() == 0) {
			return Matrix<T>();
		}
		else {
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Diag(Matrix<T> a_matrix, int a_diagonal);*/

	/**/
	/*
	NAME

		Tri - Create a Matrix instance with 1s on and below the specified diagonal, and 0s everywhere else.

	SYNOPSIS

		Matrix<T> Tri(int a_rows, int a_columns, int a_diagonal);
			a_rows -> The number of rows in the Matrix instance
			a_columns -> The number of columns in the Matrix instance
			a_diagonal -> The diagonal of 1s in the Matrix instance

	DESCRIPTION

		This function creates a Matrix instance with 1s on and below the specified diagonal, and 0s everywhere else.

	RETURNS

		A Matrix instance with 1s on and below the specified diagonal, and 0s everywhere else.
	*/
	/**/
	template <typename T>
	Matrix<T> Tri(int a_rows, int a_columns, int a_diagonal) {
		vector<vector<T>> result;
		int total_columns;
		if (a_rows < 0) {
			cout << "Error: Matrix dimensions must be positive." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		// If the user passes in a negative number or 0 for the number of columns, then we set the number of columns to the number of rows.
		if (a_columns <= 0) {
			total_columns = a_rows;
		}
		else {
			total_columns = a_columns;
		}
		for (int row_index = 0; row_index < a_rows; row_index++) {
			vector<T> row;
			for (int column_index = 0; column_index < total_columns; column_index++) {
				// If we get here, then we are on or below the diagonal
				if (column_index <= row_index + a_diagonal) {
					row.push_back(1);
				}
				else {
					row.push_back(0);
				}
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Tri(int a_rows, int a_columns, int a_diagonal);*/

	/**/
	/*
	NAME

		SaveAsCSV - Save the data in a Matrix instance to a CSV file.

	SYNOPSIS

		void SaveAsCSV(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose data will be saved to a CSV file

	DESCRIPTION

		This function asks the user for the name of the file they want to save the data to.
		After the user enters the name of the file, the data in the Matrix instance is written to the CSV file.
		To help with figuring out how to write data to a CSV file, the following links were consulted:
		https://stackoverflow.com/questions/6406356/how-to-write-vector-values-to-a-file
		https://cplusplus.com/forum/general/5898/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void SaveAsCSV(Matrix<T> a_matrix) {
		vector<vector<T>> data = a_matrix.GetData();

		// Prepare the file for writing, and user input
		ofstream output_file;
		string file_name, file_extension;
		cout << "Enter the name of the file you want to save the data to: ";
		cin >> file_name;

		// Write the data to the file
		string file_path = file_name + ".csv";
		output_file.open(file_path);
		for (int current_row = 0; current_row < data.size(); current_row++) {
			for (int current_column = 0; current_column < data[current_row].size(); current_column++) {
				output_file << data[current_row][current_column] << ",";
			}
			output_file << "\r";
		}
	
		output_file.close();
	}
	/*void SaveAsCSV(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		CopyTo - Copy the data from one Matrix instance to another Matrix instance.

	SYNOPSIS

		void CopyTo(Matrix<T>& a_destination, Matrix<T> a_source);
			a_destination -> The Matrix instance where the data will be copied to
			a_source -> The Matrix instance where the data will be copied from

	DESCRIPTION

		This function copies the data from one Matrix instance to another Matrix instance.
		We need to pass in the Matrix instances by reference so we can modify them.
		Otherwise, we would be working with copies of the Matrix instances, and not the actual Matrix instances.
		This was a problem I ran into when I was trying to implement the 'CopyTo' function. But thanks to a conversation with Microsoft Copilot,
		I was able to figure out how to implement the function correctly.

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void CopyTo(Matrix<T>& a_destination, Matrix<T> a_source) {
		// Make sure the dimensions are the same
		if ((a_destination.GetRows() == a_source.GetRows()) && (a_destination.GetCols() == a_source.GetCols())) {
			vector<vector<T>> result;
			for (int row_index = 0; row_index < a_destination.GetRows(); row_index++) {
				vector<T> row;
				for (int column_index = 0; column_index < a_destination.GetCols(); column_index++) {
					row.push_back(a_source.GetData()[row_index][column_index]);
				}
				result.push_back(row);
			}
			// Copy the data from the source Matrix instance to the destination Matrix instance
			a_destination.SetData(result);
		}
		else {
			cout << "Error: Matrix dimensions must agree." << endl;
		}
	}
	/*void CopyTo(Matrix<T>& a_destination, Matrix<T> a_source);*/

	/**/
	/*
	NAME

		Shape - Obtain the dimensions of a Matrix instance.

	SYNOPSIS

		vector<int> Shape(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose dimensions will be returned

	DESCRIPTION

		This function stores the dimensions of the Matrix instance in a vector, with the first element being the number of rows,
		and the second element being the number of columns.

	RETURNS

		A vector containing the dimensions of the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	vector<int> Shape(Matrix<T> a_matrix) {
		vector<int> dimensions = {a_matrix.GetRows(), a_matrix.GetCols()};
		return dimensions;
	}
	/*vector<int> Shape(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Flatten - Flatten a Matrix instance into a vector.

	SYNOPSIS

		vector<T> Flatten(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be flattened

	DESCRIPTION

		This function flattens a Matrix instance into a vector.

	RETURNS

		A vector that represents the flattened Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	vector<T> Flatten(Matrix<T> a_matrix) {
		vector<T> result;
		for (int row_index = 0; row_index < a_matrix.GetRows(); row_index++) {
			for (int column_index = 0; column_index < a_matrix.GetCols(); column_index++) {
				result.push_back(a_matrix.GetData()[row_index][column_index]);
			}
		}
		return result;
	}
	/*vector<T> Flatten(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Transpose - Obtain the transpose of a Matrix instance.

	SYNOPSIS

		Matrix<T> Transpose(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose Transpose will be returned

	DESCRIPTION

		This function creates a Matrix instance that is the transpose of the Matrix instance passed in as a parameter.
		To get the Transpose of the Matrix, we simply swap the rows and columns of the Matrix instance.
		To help get an idea of how to implement this function, the following link was consulted:
		https://github.com/pankajkompella/Hackerrank-SI-Basic/blob/master/Transpose%20matrix.c

	RETURNS

		A Matrix instance that is the transpose of the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<T> Transpose(Matrix<T> a_matrix) {
		// Check if the matrix is empty
		if (a_matrix.GetRows() == 0 && a_matrix.GetCols() == 0) {
			cout << "Error: Cannot take the transpose of an Empty matrix." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int column_index = 0; column_index < a_matrix.GetCols(); column_index++) {
			vector<T> row;
			for (int row_index = 0; row_index < a_matrix.GetRows(); row_index++) {
				row.push_back(a_matrix.GetData()[row_index][column_index]);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Transpose(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Concatenate - Join a vector of Matrix instances along a specified axis.

	SYNOPSIS

		Matrix<T> Concatenate(vector<Matrix<T>> a_matrices, int a_axis);
			a_matrices -> The vector of Matrix instances to be concatenated
			a_axis -> The axis along which the Matrix instances will be concatenated

	DESCRIPTION

		This function creates a Matrix instance where the Matrix instances passed in as a parameter are concatenated along the specified axis.

	RETURNS

		A Matrix instance containing the Matrix instances passed in as a parameter, concatenated along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Concatenate(vector<Matrix<T>> a_matrices, int a_axis) {
		if (a_axis == 0) {
			// If we get here, then we are concatenating the matrices along the rows
			for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
				if (current_matrix > 0) {
					if (a_matrices[current_matrix].GetCols() != a_matrices[current_matrix - 1].GetCols()) {
						cout << "Error: Number of columns be the same for axis equal to 0." << endl;
						cout << "Returning Empty matrix." << endl;
						return Matrix<T>();
					}
				}
				else {
					continue;
				}
			}
			vector<vector<T>> result;
			for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
				for (int current_row = 0; current_row < a_matrices[current_matrix].GetRows(); current_row++) {
					vector<T> row;
					for (int current_column = 0; current_column < a_matrices[current_matrix].GetCols(); current_column++) {
						row.push_back(a_matrices[current_matrix].GetData()[current_row][current_column]);
					}
					result.push_back(row);
				}
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are concatenating the matrices along the columns
			for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
				if (current_matrix > 0) {
					if (a_matrices[current_matrix].GetRows() != a_matrices[current_matrix - 1].GetRows()) {
						cout << "Error: Number of rows be the same for axis equal to 1." << endl;
						cout << "Returning Empty matrix." << endl;
						return Matrix<T>();
					}
				}
				else {
					continue;
				}
			}
			vector<vector<T>> result = a_matrices[0].GetData();
			for (int current_matrix = 1; current_matrix < a_matrices.size(); current_matrix++) {
				for (int current_row = 0; current_row < a_matrices[current_matrix].GetRows(); current_row++) {
					for (int current_column = 0; current_column < a_matrices[current_matrix].GetCols(); current_column++) {
						T current_value = a_matrices[current_matrix].GetData()[current_row][current_column];
						result[current_row].push_back(current_value);
					}
				}
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we just flatten the matrices into a single vector
			vector<vector<T>> result;
			vector<T> row;
			for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
				for (int current_row = 0; current_row < a_matrices[current_matrix].GetRows(); current_row++) {
					for (int current_column = 0; current_column < a_matrices[current_matrix].GetCols(); current_column++){
						T value = a_matrices[current_matrix].GetData()[current_row][current_column];
						row.push_back(value);
					}
				}
			}
			result.push_back(row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Concatenate(vector<Matrix<T>> a_matrices, int a_axis);*/

	/**/
	/*
	NAME

		HStack - Stack a vector of Matrix instances horizontally.

	SYNOPSIS

		Matrix<T> HStack(vector<Matrix<T>> a_matrices);
			a_matrices -> The vector of Matrix instances to be stacked horizontally

	DESCRIPTION

		This function creates a Matrix instance where the Matrix instances passed in as a parameter are stacked horizontally (column-wise).

	RETURNS

		A new Matrix instance where all of the Matrix instances passed in as a parameter are stacked horizontally (column-wise).
	*/
	/**/
	template <typename T>
	Matrix<T> HStack(vector<Matrix<T>> a_matrices) {
		return Concatenate<T>(a_matrices, 1);
	}
	/*Matrix<T> HStack(vector<Matrix<T>> a_matrices);*/

	/**/
	/*
	NAME

		VStack - Stack a vector of Matrix instances vertically.

	SYNOPSIS

		Matrix<T> VStack(vector<Matrix<T>> a_matrices);
			a_matrices -> The vector of Matrix instances to be stacked vertically

	DESCRIPTION

		This function creates a Matrix instance where the Matrix instances passed in as a parameter are stacked vertically (row-wise).

	RETURNS

		A new Matrix instance where all of the Matrix instances passed in as a parameter are stacked vertically (row-wise).
	*/
	/**/
	template <typename T>
	Matrix<T> VStack(vector<Matrix<T>> a_matrices) {
		return Concatenate<T>(a_matrices, 0);
	}
	/*Matrix<T> VStack(vector<Matrix<T>> a_matrices);*/

	/**/
	/*
	NAME

		Column_Stack - Stack 1D Matrix instances as columns into a 2D Matrix instance.

	SYNOPSIS

		Matrix<T> Column_Stack(vector<Matrix<T>> a_matrices);
			a_matrices -> The vector of Matrix instances to be stacked as columns

	DESCRIPTION

		This function creates a new Matrix instance where the Matrix instances passed in as a parameter are stacked as columns,
		depending on whether the instances passed in are 1D or 2D.

	RETURNS

		A new Matrix instance where all of the Matrix instances passed in as a parameter are stacked as columns.
	*/
	/**/
	template <typename T>
	Matrix<T> Column_Stack(vector<Matrix<T>> a_matrices) {
		bool all_1d = true;
		for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
			if (a_matrices[current_matrix].GetRows() != 1) {
				all_1d = false;
				break;
			}
			else {
				continue;
			}
		}
		if (all_1d == true) {
			// If we get here, then we are dealing with 1D Matrix instances, so we need to transpose them
			vector<Matrix<T>> temp;
			for (int current_matrix = 0; current_matrix < a_matrices.size(); current_matrix++) {
				temp.push_back(Transpose<T>(a_matrices[current_matrix]));
			}
			return HStack<T>(temp);
		}
		else {
			// If we get here, then we are dealing with 2D Matrix instances, so we can just call the 'HStack' function
			return HStack<T>(a_matrices);
		}
	}
	/*Matrix<T> Column_Stack(vector<Matrix<T>> a_matrices);*/

	/**/
	/*
	NAME

		Row_Stack - Stack a vector of Matrix instances in sequence vertically (row-wise).

	SYNOPSIS

		Matrix<T> Row_Stack(vector<Matrix<T>> a_matrices);
			a_matrices -> The vector of Matrix instances to be stacked vertically

	DESCRIPTION

		This function creates a Matrix instance where the Matrix instances passed in as a parameter are stacked vertically.
		To do this, we simply call the 'VStack' function.

	RETURNS

		A new Matrix instance where all of the Matrix instances passed in as a parameter are stacked vertically.
	*/
	/**/
	template <typename T>
	Matrix<T> Row_Stack(vector<Matrix<T>> a_matrices) {
		return VStack<T>(a_matrices);
	}
	/*Matrix<T> Row_Stack(vector<Matrix<T>> a_matrices);*/

	/**/
	/*
	NAME

		Unique - Find the unique elements of a Matrix instance.

	SYNOPSIS

		vector<T> Unique(Matrix<T> a_matrix, bool a_unique_indices, bool a_unique_counts);
			a_matrix -> The Matrix instance whose unique elements will be returned
			a_unique_indices -> Whether or not to return the indices of the unique elements
			a_unique_counts -> Whether or not to return the counts of the unique elements

	DESCRIPTION

		This function stores the unique elements of a Matrix instance in a vector. In addition, it also
		stores the indices of the unique elements, as well as how many times those unique elements
		occur in the Matrix, depending on the values of the 'a_unique_indices' and 'a_unique_counts' parameters, respectively.
		To help with figuring out how to implement this function, the following links were consulted:
		https://stackoverflow.com/questions/571394/how-to-find-out-if-an-item-is-present-in-a-stdvector
		https://www.geeksforgeeks.org/sort-c-stl/

	RETURNS

		A vector containing the unique elements of the Matrix instance passed in as a parameter, sorted in ascending order.
	*/
	/**/
	template <typename T>
	vector<T> Unique(Matrix<T> a_matrix, bool a_unique_indices, bool a_unique_counts) {
		vector<T> result, unique_indices, unique_counts;

		// Store the unique elements in a vector
		for (int row_index = 0; row_index < a_matrix.GetRows(); row_index++) {
			for (int column_index = 0; column_index < a_matrix.GetCols(); column_index++) {
				T current_value = a_matrix.GetData()[row_index][column_index];
				if (find(result.begin(), result.end(), current_value) != result.end()) {
					continue;
				}
				else {
					result.push_back(current_value);
				}
			}
		}

		// Store the indices of the unique elements in another vector
		if (a_unique_indices == true) {
			vector<T> matrix_data = Flatten(a_matrix), temp;
			for (int current_index = 0; current_index < matrix_data.size(); current_index++) {
				if (find(temp.begin(), temp.end(), matrix_data[current_index]) != temp.end()) {
					continue;
				}
				else {
					temp.push_back(matrix_data[current_index]);
					unique_indices.push_back(current_index);
				}
			}
			for (int current_index = 0; current_index < unique_indices.size(); current_index++) {
				cout << unique_indices[current_index] << endl;
			}
		}

		// Store the counts of the unique elements in another vector
		if (a_unique_counts == true) {
			for (int current_index = 0; current_index < result.size(); current_index++) {
				int count = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
						if (a_matrix.GetData()[current_row][current_column] == result[current_index]) {
							count++;
						}
						else {
							continue;
						}
					}
				}
				unique_counts.push_back(count);
			}
			for (int current_index = 0; current_index < unique_counts.size(); current_index++) {
				cout << unique_counts[current_index] << endl;
			}
		}

		sort(result.begin(), result.end());
		return result;
	}
	/*vector<T> Unique(Matrix<T> a_matrix, bool a_unique_indices, bool a_unique_counts);*/

	/**/
	/*
	NAME

		Flip - flip the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		Matrix<T> Flip(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance to be flipped
			a_axis -> The axis along which the Matrix instance will be flipped

	DESCRIPTION

		This function flips the elements of a Matrix instance along a specified axis.
		To help with initializing the 2D vector that will store the result, the following link was consulted:
		https://www.geeksforgeeks.org/2d-vector-in-cpp-with-user-defined-size/
		And to help with figuring out how to flip the Matrix instance along the specified axis, a conversation with Microsoft Copilot was had.
		This conversation led to the following link being consulted: https://stackoverflow.com/questions/47930428/how-to-rotate-an-array-by-%c2%b1-180-in-an-efficient-way

	RETURNS

		A new Matrix instance with the elements of the original Matrix instance flipped along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Flip(Matrix<T> a_matrix, int a_axis) {
		// Check if the matrix is empty
		if (a_matrix.GetRows() == 0 && a_matrix.GetCols() == 0) {
			cout << "Error: Cannot flip an empty matrix." << endl;
			cout << "Returning empty matrix." << endl;
			return Matrix<T>();
		}

		// First, we have to initialize the 2D vector that will store the result
		vector<vector<T>> result(a_matrix.GetRows());
		for (int current_vector = 0; current_vector < a_matrix.GetRows(); current_vector++) {
			result[current_vector] = vector<T>(a_matrix.GetCols());
		}

		// Then, we have to determine how the Matrix instance will be flipped based on the value of 'a_axis'
		if (a_axis == 0) {
			// If we get here, then we are flipping the Matrix instance vertically
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					result[a_matrix.GetRows() - 1 - current_row][current_column] = a_matrix.GetData()[current_row][current_column];
				}
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are flipping the Matrix instance horizontally
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					result[current_row][a_matrix.GetCols() - 1 - current_column] = a_matrix.GetData()[current_row][current_column];
				}
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then an invalid value was passed in for 'a_axis'
			vector<T> data = Flatten(a_matrix);
			reverse(data.begin(), data.end());
			int current_element = 0;
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					result[current_row][current_column] = data[current_element];
					current_element++;
				}
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Flip(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Repeat - Repeat the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		Matrix<T> Repeat(Matrix<T> a_matrix, int a_repeats, int a_axis);
			a_matrix -> The Matrix instance whose elements will be repeated
			a_repeats -> The number of times the elements of the Matrix instance will be repeated
			a_axis -> The axis along which the elements of the Matrix instance will be repeated

	DESCRIPTION

		This function repeats the elements of a Matrix instance along a specified axis.

	RETURNS

		A new Matrix instance with the elements of the original Matrix instance repeated the specified number of times
		along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Repeat(Matrix<T> a_matrix, int a_repeats, int a_axis) {
		// Check if the matrix is empty
		if (a_matrix.GetRows() == 0 && a_matrix.GetCols() == 0) {
			cout << "Error: Cannot repeat an empty matrix." << endl;
			cout << "Returning empty matrix." << endl;
			return Matrix<T>();
		}
		vector<vector<T>> data = a_matrix.GetData(), result;

		// First, we have to check the value of 'a_repeats'
		if (a_repeats < 0) {
			cout << "Error: Number of repeats must be positive." << endl;
			cout << "Returning empty matrix." << endl;
			return Matrix<T>();
		}
		else if (a_repeats == 0) {
			return Matrix<T>();
		}

		// Now, we have to determine how the elements of the Matrix instance will be repeated based on the value of 'a_axis'
		if (a_axis == 0) {
			// If we get here, then we are repeating the elements of the Matrix instance row-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row_to_push_back = data[current_row];
				for (int current_repeat = 0; current_repeat < a_repeats; current_repeat++) {
					result.push_back(row_to_push_back);
				}
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are repeating the elements of the Matrix instance column-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					T current_value = data[current_row][current_column];
					for (int current_repeat = 0; current_repeat < a_repeats; current_repeat++) {
						row.push_back(current_value);
					}
				}
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are repeating the elements of the Matrix instance in a flattened manner
			vector<T> flattened_data = Flatten(a_matrix), row;
			for (int current_element = 0; current_element < flattened_data.size(); current_element++) {
				T current_value = flattened_data[current_element];
				for (int current_repeat = 0; current_repeat < a_repeats; current_repeat++) {
					row.push_back(current_value);
				}
			}
			result.push_back(row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Repeat(Matrix<T> a_matrix, int a_repeats, int a_axis);*/

	/**/
	/*
	NAME

		Trim_Zeros - Remove leading and/or trailing zeros from a Matrix instance.

	SYNOPSIS

		Matrix<T> Trim_Zeros(Matrix<T> a_matrix, string a_trim);
			a_matrix -> The Matrix instance to be trimmed
			a_trim -> Whether to remove the leading zeros (represented by 'f' and 'F'), the trailing zeros (represented by 'b' and 'B'),
					  or both the leading and trailing zeros (represented by 'fb' and 'FB')

	DESCRIPTION

		This function removes leading and/or trailing zeros from a Matrix instance, depending on the value of 'a_trim'.

	RETURNS

		A Matrix instance with the leading and/or trailing zeros removed, based on the value of 'a_trim'.
	*/
	/**/
	template <typename T>
	Matrix<T> Trim_Zeros(Matrix<T> a_matrix, string a_trim) {
		// Make sure the Matrix is 1D, since the original NumPy function only works with 1D arrays
		if (a_matrix.GetRows() > 1) {
			cout << "Error: Matrix must be 1D." << endl;
			cout << "Returning Empty matrix." << endl;
			return Matrix<T>();
		}
		else {
			vector<vector<T>> result;
			vector<T> row = a_matrix.GetData()[0];
			int leading_zero_count = 0, trailing_zero_count = 0;
			// First, we determine how many leading and trailing zeros there are
			for (int current_element = 0; current_element < row.size(); current_element++) {
				if (row[current_element] == 0) {
					leading_zero_count++;
				}
				else {
					break;
				}
			}
			for (int current_element = row.size() - 1; current_element >= 0; current_element--) {
				if (row[current_element] == 0) {
					trailing_zero_count++;
				}
				else {
					break;
				}
			}

			// Then, we remove the leading and/or trailing zeros based on the value of 'a_trim'
			if (a_trim == "f" || a_trim == "F") {
				row.erase(row.begin(), row.begin() + leading_zero_count);
				result.push_back(row);
			}
			else if (a_trim == "b" || a_trim == "B") {
				row.erase(row.end() - trailing_zero_count, row.end());
				result.push_back(row);
			}
			else if (a_trim == "fb" || a_trim == "FB") {
				row.erase(row.begin(), row.begin() + leading_zero_count);
				row.erase(row.end() - trailing_zero_count, row.end());
				result.push_back(row);
			}
			else {
				// If we get here, then we are dealing with an invalid value for 'a_trim'
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Trim_Zeros(Matrix<T> a_matrix, string a_trim);*/

	/**/
	/*
	NAME

		Delete - remove elements from a Matrix instance at a specified index

	SYNOPSIS

		void Delete(Matrix<T>& a_matrix, int a_index, int a_axis);
			a_matrix -> The Matrix instance that will be passed in as a reference, so we can directly change its attributes
			a_index -> The index at which the corresponding elements will be deleted
			a_axis -> The axis along which the elements will be deleted

	DESCRIPTION

		This function removes elements from a Matrix instance, depending on the values of "a_index" and "a_axis".

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Delete(Matrix<T>& a_matrix, int a_index, int a_axis) {
		if (a_index < 0) {
			cout << "Error: Invalid index value detected." << endl;
			return;
		}
		else {
			vector<vector<T>> matrix_data = a_matrix.GetData();
			int current_column_amount = a_matrix.GetCols(), current_row_amount = a_matrix.GetRows();
			if (a_axis == 0) {
				// If we get here, then we are deleting rows
				matrix_data.erase(matrix_data.begin() + a_index);
				a_matrix.SetRows(current_row_amount - 1);
				a_matrix.SetData(matrix_data);
			}
			else if (a_axis == 1) {
				// If we get here, then we are deleting columns
				vector<vector<T>> temp;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					vector<T> row;
					for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
						T current_element = matrix_data[current_row][current_column];
						if (current_column != a_index) {
							row.push_back(current_element);
						}
						else {
							continue;
						}
					}
					temp.push_back(row);
				}
				a_matrix.SetCols(current_column_amount - 1);
				a_matrix.SetData(temp);
			}
			else {
				// If we get here, we simply flatten the matrix, then remove the element at the specified index
				vector<vector<T>> temp;
				vector<T> flattened_data = Flatten(a_matrix);
				flattened_data.erase(flattened_data.begin() + a_index);
				temp.push_back(flattened_data);
				a_matrix.SetData(temp);
				a_matrix.SetRows(1);
				a_matrix.SetCols(flattened_data.size());
			}
		}
	}
	/*void Delete(Matrix<T>& a_matrix, int a_index, int a_axis);*/

	/**/
	/*
	NAME

		Insert - insert an element into a Matrix instance at a specified index

	SYNOPSIS

		void Insert(Matrix<T>& a_matrix, int a_index, T a_value, int a_axis);
			a_matrix -> The Matrix instance that will be passed in as a reference, so we can directly change its attributes
			a_index -> The index at which the element will be inserted
			a_value -> The value of the element to be inserted
			a_axis -> The axis along which the element will be inserted

	DESCRIPTION

		This function removes elements from a Matrix instance, depending on the values of "a_index" and "a_axis".

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Insert(Matrix<T>& a_matrix, int a_index, T a_value, int a_axis) {
		if (a_index < 0) {
			cout << "Error: Invalid index value detected." << endl;
			return;
		}
		else {
			if (a_axis == 0) {
				// If we get here, then we are inserting a row
				vector<vector<T>> matrix_data = a_matrix.GetData();
				vector<T> row(a_matrix.GetCols(), a_value);
				matrix_data.insert(matrix_data.begin() + a_index, row);
				a_matrix.SetData(matrix_data);
				a_matrix.SetRows(a_matrix.GetRows() + 1);
			}
			else if (a_axis == 1) {
				// If we get here, then we are inserting a column
				vector<vector<T>> matrix_data = a_matrix.GetData();
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					matrix_data[current_row].insert(matrix_data[current_row].begin() + a_index, a_value);
				}
				a_matrix.SetData(matrix_data);
				a_matrix.SetCols(a_matrix.GetCols() + 1);
			}
			else {
				// If we get here, we simply flatten the matrix, then insert the element at the specified index
				vector<T> flattened_data = Flatten(a_matrix);
				flattened_data.insert(flattened_data.begin() + a_index, a_value);
				vector<vector<T>> temp;
				temp.push_back(flattened_data);
				a_matrix.SetData(temp);
				a_matrix.SetRows(1);
				a_matrix.SetCols(flattened_data.size());
			}
		}
	}
	/*void Insert(Matrix<T>& a_matrix, int a_index, T a_value, int a_axis);*/

	/**/
	/*
	NAME

		Append - append elements to the end of a Matrix instance

	SYNOPSIS

		void Append(Matrix<T>& a_matrix_one, Matrix<T> a_matrix_two, int a_axis);
			a_matrix_one -> The Matrix instance to which elements will be appended
			a_matrix_two -> The Matrix instance containing the elements that will be appended
			a_axis -> The axis along which the elements will be appended

	DESCRIPTION

		This function takes the elements from the second Matrix instance, and appends them to
		the end of the first Matrix instance, depending on the value of 'a_axis'.
		To help get an idea of how to append the elements when 'a_axis' does not equal 0, a conversation
		with Microsoft Copilot was consulted.

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Append(Matrix<T>& a_matrix_one, Matrix<T> a_matrix_two, int a_axis) {
		if (a_axis == 0) {
			// If we get here, then we are trying to append elements to the end of each column
			if (a_matrix_two.GetCols() != a_matrix_one.GetCols()) {
				cout << "Error: Number of columns must be the same for axis 0." << endl;
				return;
			}
			else {
				vector<vector<T>> first_matrix_data = a_matrix_one.GetData(), second_matrix_data = a_matrix_two.GetData();
				for (int current_row = 0; current_row < a_matrix_two.GetRows(); current_row++) {
					vector<T> temp = second_matrix_data[current_row];
					first_matrix_data.push_back(temp);
				}
				a_matrix_one.SetData(first_matrix_data);
				a_matrix_one.SetRows(a_matrix_one.GetRows() + a_matrix_two.GetRows());
			}
		}
		else if (a_axis == 1) {
			// If we get here, then we are trying to append elements to the end of each row
			if (a_matrix_two.GetRows() != a_matrix_one.GetRows()) {
				cout << "Error: Number of rows must be the same for axis 1." << endl;
				return;
			}
			else {
				vector<vector<T>> first_matrix_data = a_matrix_one.GetData(), second_matrix_data = a_matrix_two.GetData();
				for (int current_row = 0; current_row < a_matrix_two.GetRows(); current_row++) {
					first_matrix_data[current_row].insert(first_matrix_data[current_row].end(), second_matrix_data[current_row].begin(), second_matrix_data[current_row].end());
				}
				a_matrix_one.SetData(first_matrix_data);
				a_matrix_one.SetCols(a_matrix_one.GetCols() + a_matrix_two.GetCols());
			}
		}
		else {
			// If we get here, then we simply flatten both matrices into vectors, and combine them together
			vector<vector<T>> temp;
			vector<T> first_matrix_data = Flatten(a_matrix_one), second_matrix_data = Flatten(a_matrix_two);
			first_matrix_data.insert(first_matrix_data.end(), second_matrix_data.begin(), second_matrix_data.end());
			temp.push_back(first_matrix_data);
			a_matrix_one.SetData(temp);
			a_matrix_one.SetCols(first_matrix_data.size());
			a_matrix_one.SetRows(1);
		}
	}
	/*void Append(Matrix<T>& a_matrix_one, Matrix<T> a_matrix_two, int a_axis);*/

	/**/
	/*
	NAME

		Logical_And - Perform a logical AND operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Logical_And(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '&&' operator, which performs a logical AND operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the logical AND operation performed on the two Matrix instances passed in as parameters.
	*/
	/**/
	template <typename T>
	Matrix<bool> Logical_And(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 && a_matrix2;
	}
	/*Matrix<bool> Logical_And(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Logical_Or - Perform a logical OR operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Logical_Or(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '||' operator, which performs a logical OR operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the logical OR operation performed on the two Matrix instances passed in as parameters.
	*/
	/**/
	template <typename T>
	Matrix<bool> Logical_Or(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 || a_matrix2;
	}
	/*Matrix<bool> Logical_Or(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Logical_XOR - Perform a logical XOR operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Logical_XOR(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '^' operator, which performs a logical XOR operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the logical XOR operation performed on the two Matrix instances passed in as parameters.
	*/
	/**/
	template <typename T>
	Matrix<bool> Logical_XOR(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 ^ a_matrix2;
	}
	/*Matrix<bool> Logical_XOR(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Logical_Not - Perform a logical NOT operation on a Matrix instance

	SYNOPSIS

		Matrix<bool> Logical_Not(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '!' operator, which performs a logical NOT operation on a Matrix instance.

	RETURNS

		A Matrix instance containing the result of the logical NOT operation performed on the Matrix instance passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Logical_Not(Matrix<T> a_matrix) {
		return !(a_matrix);
	}
	/*Matrix<bool> Logical_Not(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Greater - Perform a greater than operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Greater(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '>' operator, which performs a greater than operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the greater than operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Greater(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 > a_matrix2;
	}
	/*Matrix<bool> Greater(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Greater_Equal - Perform a greater than or equal to operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Greater_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '>=' operator, which performs a greater than or equal to operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the greater than or equal to operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Greater_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 >= a_matrix2;
	}
	/*Matrix<bool> Greater_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Less - Perform a less than operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Less(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '<' operator, which performs a less than operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the less than operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Less(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 < a_matrix2;
	}
	/*Matrix<bool> Less(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Less_Equal - Perform a less than or equal to operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Less_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '<=' operator, which performs a less than or equal to operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the less than or equal to operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Less_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 <= a_matrix2;
	}
	/*Matrix<bool> Less_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Equal - Perform an equal to operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '==' operator, which performs an equal to operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the equal to operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 == a_matrix2;
	}
	/*Matrix<bool> Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Not_Equal - Perform a not equal to operation on two Matrix instances.

	SYNOPSIS

		Matrix<bool> Not_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used
			a_matrix2 -> The second Matrix instance to be used

	DESCRIPTION

		This function serves as a wrapper for the overloaded '!=' operator, which performs a not equal to operation on two Matrix instances.

	RETURNS

		A Matrix instance containing the result of the not equal to operation performed on the Matrix instances passed in as a parameter.
	*/
	/**/
	template <typename T>
	Matrix<bool> Not_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		return a_matrix1 != a_matrix2;
	}
	/*Matrix<bool> Not_Equal(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Is_Finite - Determine whether or not a Matrix instance contains finite values.

	SYNOPSIS

		Matrix<bool> Is_Finite(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function determines whether or not a Matrix instance contains finite values (in other words, values that are not infinity and not nan).
		To help with figuring out how to test for nan, a conversation was had with Microsoft Copilot which led to the following link being consulted:
		https://www.codeproject.com/Tips/676287/C-and-Cplusplus-Use-isnan-to-Test-for-nan

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is finite.
	*/
	/**/
	template <typename T>
	Matrix<bool> Is_Finite(Matrix<T> a_matrix) {
		vector<vector<bool>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<bool> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T current_value = a_matrix.GetData()[current_row][current_column];
				if (current_value == g_inf || current_value == g_ninf ||isnan(current_value)) {
					// If we get here, then the current value is not finite
					row.push_back(false);
				}
				else {
					row.push_back(true);
				}
			}
			result.push_back(row);
		}
		return Matrix<bool>(result);
	}
	/*Matrix<bool> Is_Finite(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Is_Inf - Determine whether or not a Matrix instance contains positive or negative infinity values.

	SYNOPSIS

		Matrix<bool> Is_Inf(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function determines whether or not a Matrix instance contains positive or negative infinity values.

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is
		positive or negative infinity.
	*/
	/**/
	template <typename T>
	Matrix<bool> Is_Inf(Matrix<T> a_matrix) {
		vector<vector<bool>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<bool> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T current_value = a_matrix.GetData()[current_row][current_column];
				if (current_value == g_inf || current_value == g_ninf) {
					// If we get here, then the current value is either positive or negative infinity
					row.push_back(true);
				}
				else {
					row.push_back(false);
				}
			}
			result.push_back(row);
		}
		return Matrix<bool>(result);
	}
	/*Matrix<bool> Is_Inf(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Is_Nan - Determine whether or not a Matrix instance contains nan values.

	SYNOPSIS

		Matrix<bool> Is_Nan(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function determines whether or not a Matrix instance contains nan values.

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is nan.
	*/
	/**/
	template <typename T>
	Matrix<bool> Is_Nan(Matrix<T> a_matrix) {
		vector<vector<bool>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<bool> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T current_value = a_matrix.GetData()[current_row][current_column];
				if (isnan(current_value)) {
					// If we get here, then the current value is nan
					row.push_back(true);
				}
				else {
					row.push_back(false);
				}
			}
			result.push_back(row);
		}
		return Matrix<bool>(result);
	}
	/*Matrix<bool> Is_Nan(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Contains - Determine whether or not a Matrix instance contains a specified value.

	SYNOPSIS

		Matrix<bool> Contains(Matrix<T> a_matrix, T a_value);
			a_matrix -> The Matrix instance to be used
			a_value -> The value to be searched for

	DESCRIPTION

		This function searches a Matrix instance for a specified value, and returns a Matrix instance of boolean values, where each value represents
		whether or not the corresponding value in the original Matrix instance is equal to the specified value.

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is
		equal to the specified value.
	*/
	/**/
	template <typename T>
	Matrix<bool> Contains(Matrix<T> a_matrix, T a_value) {
		vector<vector<bool>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<bool> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T current_value = a_matrix.GetData()[current_row][current_column];
				if (current_value == a_value) {
					// If we get here, then the current value is equal to the specified value
					row.push_back(true);
				}
				else {
					row.push_back(false);
				}
			}
			result.push_back(row);
		}
		return Matrix<bool>(result);
	}
	/*Matrix<bool> Contains(Matrix<T> a_matrix, T a_value);*/

	/**/
	/*
	NAME

		Is_NegInf - Determine whether or not a Matrix instance contains negative infinity values.

	SYNOPSIS

		Matrix<bool> Is_NegInf(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function traverses a Matrix instance, and determines whether or not it contains negative infinity values.

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is
		equal to negative infinity.
	*/
	/**/
	template <typename T>
	Matrix<bool> Is_NegInf(Matrix<T> a_matrix) {
		return Contains(a_matrix, g_ninf);
	}
	/*Matrix<bool> Is_NegInf(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Is_PosInf - Determine whether or not a Matrix instance contains positive infinity values.

	SYNOPSIS

		Matrix<bool> Is_PosInf(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance to be used

	DESCRIPTION

		This function traverses a Matrix instance, and determines whether or not it contains positive infinity values.

	RETURNS

		A Matrix instance of boolean values, where each value represents whether or not the corresponding value in the original Matrix instance is
		equal to positive infinity.
	*/
	/**/
	template <typename T>
	Matrix<bool> Is_PosInf(Matrix<T> a_matrix) {
		return Contains(a_matrix, g_inf);
	}
	/*Matrix<bool> Is_PosInf(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Sum - Calculate the sum of the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		vector<T> Sum(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be summed
			a_axis -> The axis along which the elements will be summed

	DESCRIPTION

		This function computes the sum of the elements of a Matrix instance along a specified axis.

	RETURNS

		A vector containing the sum of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	vector<T> Sum(Matrix<T> a_matrix, int a_axis) {
		vector<T> result;
		if (a_axis == 0) {
			// If we get here, then we are summing the elements of the Matrix instance column-wise
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T sum = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					sum += a_matrix.GetData()[current_row][current_column];
				}
				result.push_back(sum);
			}
		}
		else if (a_axis == 1) {
			// If we get here, then we are summing the elements of the Matrix instance row-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				T sum = 0;
				for (int current_column = 0; current_column < a_matrix.GetRows(); current_column++) {
					sum += a_matrix.GetData()[current_row][current_column];
				}
				result.push_back(sum);
			}
		}
		else{
			// If we get here, then we are summing the elements of the Matrix instance in a flattened manner
			vector<T> temp = Flatten(a_matrix);
			T sum = 0;
			for (int current_element = 0; current_element < temp.size(); current_element++) {
				sum += temp[current_element];
			}
			result.push_back(sum);
		}
		return result;
	}
	/*vector<T> Sum(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Prod - Calculate the product of the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		vector<T> Prod(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be multiplied
			a_axis -> The axis along which the elements will be multiplied

	DESCRIPTION

		This function computes the product of the elements of a Matrix instance along a specified axis.

	RETURNS

		A vector containing the product of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	vector<T> Prod(Matrix<T> a_matrix, int a_axis) {
		vector<T> result;
		if (a_axis == 0) {
			// If we get here, then we are multiplying the elements of the Matrix instance column-wise
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T product = 1;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					product *= a_matrix.GetData()[current_row][current_column];
				}
				result.push_back(product);
			}
		}
		else if (a_axis == 1) {
			// If we get here, then we are multiplying the elements of the Matrix instance row-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				T product = 1;
				for (int current_column = 0; current_column < a_matrix.GetRows(); current_column++) {
					product *= a_matrix.GetData()[current_row][current_column];
				}
				result.push_back(product);
			}
		}
		else {
			// If we get here, then we are multiplying the elements of the Matrix instance in a flattened manner
			vector<T> temp = Flatten(a_matrix);
			T product = 1;
			for (int current_element = 0; current_element < temp.size(); current_element++) {
				product *= temp[current_element];
			}
			result.push_back(product);
		}
		return result;
	}
	/*vector<T> Prod(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanSum - Calculate the sum of the elements of a Matrix instance along a specified axis, treating nan values as 0.

	SYNOPSIS

		vector<T> NanSum(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be summed
			a_axis -> The axis along which the elements will be summed

	DESCRIPTION

		This function computes the sum of the elements of a Matrix instance along a specified axis,
		after converting all nan values to 0.

	RETURNS

		A vector containing the sum of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	vector<T> NanSum(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> temp = a_matrix.GetData();
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// If we get here, then we are dealing with a nan value
				if (isnan(temp[current_row][current_column])) {
					temp[current_row][current_column] = 0;
				}
			}
		}
		a_matrix.SetData(temp);
		return Sum(a_matrix, a_axis);
	}
	/*vector<T> NanSum(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanProd - Calculate the product of the elements of a Matrix instance along a specified axis, treating nan values as 1.

	SYNOPSIS

		vector<T> NanProd(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be multiplied
			a_axis -> The axis along which the elements will be multiplied

	DESCRIPTION

		This function computes the product of the elements of a Matrix instance along a specified axis,
		after converting all nan values to 1.

	RETURNS

		A vector containing the product of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	vector<T> NanProd(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> temp = a_matrix.GetData();
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// If we get here, then the current value is nan
				if (isnan(temp[current_row][current_column])) {
					temp[current_row][current_column] = 1;
				}
			}
		}
		a_matrix.SetData(temp);
		return Prod(a_matrix, a_axis);
	}
	/*vector<T> NanProd(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		CumSum - Calculate the cumulative sum of the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		Matrix<T> CumSum(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be summed
			a_axis -> The axis along which the elements will be summed

	DESCRIPTION

		This function computes the cumulative sum of the elements of a Matrix instance along a specified axis.

	RETURNS

		A Matrix instance containing the cumulative sum of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> CumSum(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> result;
		if (a_axis == 1) {
			// If we get here, then we are summing the elements of the Matrix instance row-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				int sum = 0;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					sum += a_matrix.GetData()[current_row][current_column];
					row.push_back(sum);
				}
				result.push_back(row);
			}
		}
		else if (a_axis == 0) {
			// If we get here, then we are summing the elements of the Matrix instance column-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (current_row == 0) {
						row.push_back(a_matrix.GetData()[current_row][current_column]);
					}
					else {
						T sum = a_matrix.GetData()[current_row][current_column] + result[current_row - 1][current_column];
						row.push_back(sum);
					}
				}
				result.push_back(row);
			}
		}
		else {
			// If we get here, then we are summing the elements of the Matrix instance in a flattened manner
			vector<T> row;
			int sum = 0;
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					sum += a_matrix.GetData()[current_row][current_column];
					row.push_back(sum);
				}
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> CumSum(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		CumProd - Calculate the cumulative product of the elements of a Matrix instance along a specified axis.

	SYNOPSIS

		Matrix<T> CumProd(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be multiplied
			a_axis -> The axis along which the elements will be multiplied

	DESCRIPTION

		This function computes the cumulative product of the elements of a Matrix instance along a specified axis.

	RETURNS

		A Matrix instance containing the cumulative product of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> CumProd(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> result;
		if (a_axis == 1) {
			// If we get here, then we are multiplying the elements of the Matrix instance row-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				T product = 1;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					product *= a_matrix.GetData()[current_row][current_column];
					row.push_back(product);
				}
				result.push_back(row);
			}
		}
		else if (a_axis == 0) {
			// If we get here, then we are multiplying the elements of the Matrix instance column-wise
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (current_row == 0) {
						row.push_back(a_matrix.GetData()[current_row][current_column]);
					}
					else {
						T product = a_matrix.GetData()[current_row][current_column] * result[current_row - 1][current_column];
						row.push_back(product);
					}
				}
				result.push_back(row);
			}
		}
		else {
			// If we get here, then we are multiplying the elements of the Matrix instance in a flattened manner
			vector<T> row;
			T product = 1;
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					product *= a_matrix.GetData()[current_row][current_column];
					row.push_back(product);
				}
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> CumProd(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanCumSum - Calculate the cumulative sum of the elements of a Matrix instance along a specified axis, treating nan values as 0.

	SYNOPSIS

		Matrix<T> NanCumSum(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be summed
			a_axis -> The axis along which the elements will be summed

	DESCRIPTION

		This function computes the cumulative sum of the elements of a Matrix instance along a specified axis,
		after converting all nan values to 0.

	RETURNS

		A Matrix instance containing the cumulative sum of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> NanCumSum(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> temp = a_matrix.GetData();
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// If we get here, then the current value is nan
				if (isnan(temp[current_row][current_column])) {
					temp[current_row][current_column] = 0;
				}
			}
		}
		a_matrix.SetData(temp);
		return CumSum(a_matrix, a_axis);
	}
	/*Matrix<T> NanCumSum(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanCumProd - Calculate the cumulative product of the elements of a Matrix instance along a specified axis, treating nan values as 1.

	SYNOPSIS

		Matrix<T> NanCumProd(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance whose elements will be multiplied
			a_axis -> The axis along which the elements will be multiplied

	DESCRIPTION

		This function computes the cumulative product of the elements of a Matrix instance along a specified axis,
		after converting all nan values to 1.

	RETURNS

		A Matrix instance containing the cumulative product of the elements of the Matrix instance along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> NanCumProd(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> temp = a_matrix.GetData();
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// If we get here, then the current value is nan
				if (isnan(temp[current_row][current_column])) {
					temp[current_row][current_column] = 1;
				}
			}
		}
		a_matrix.SetData(temp);
		return CumProd(a_matrix, a_axis);
	}
	/*Matrix<T> NanCumProd(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Power - Raise a base to the power of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Power(Matrix<T> a_matrix, T a_base);
			a_matrix -> The Matrix instance whose elements will be used as exponents
			a_base -> The base to be raised to the power of the elements of the Matrix instance

	DESCRIPTION

		This function raises a base to the power of the elements of a Matrix instance.

	RETURNS

		A Matrix instance containing the result of raising the base to the power of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Power(Matrix<T> a_matrix, T a_base) {
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are raising the base to the power of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];

				/*
				Make sure that the current value is not less than or equal to 0.
				Otherwise, we will be end up with undefined numbers.
				*/
				if (a_base == 0 && current_value <= 0) {
					cout << "Error: Cannot raise 0 to the power of 0 or less." << endl;
					return Matrix<T>();
				}

				row.push_back(pow(a_base, current_value));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Power(Matrix<T> a_matrix, T a_base);*/

	/**/
	/*
	NAME

		Exp - Raise Euler's constant to the power of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Exp(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose elements will be used as exponents for Euler's constant

	DESCRIPTION

		This function raises Euler's constant to the power of the elements of a Matrix instance.

	RETURNS

		A Matrix instance containing the result of raising Euler's constant to the power of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Exp(Matrix<T> a_matrix) {
		// To avoid a type conversion error, we cast Euler's constant to type T
		T base = (T) g_euler_constant;
		return Power(a_matrix, base);
	}
	/*Matrix<T> Exp(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Exp2 - Raise 2 to the power of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Exp2(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose elements will be used as exponents for 2

	DESCRIPTION

		This function raises 2 to the power of the elements of a Matrix instance.

	RETURNS

		A Matrix instance containing the result of raising 2 to the power of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Exp2(Matrix<T> a_matrix) {
		// To avoid a type conversion error, we cast 2 to type T
		T base = (T) 2;
		return Power(a_matrix, base);
	}
	/*Matrix<T> Exp2(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		ExpM1 - Raise Euler's constant to the power of the elements of a Matrix instance, and subtract 1 from the result.

	SYNOPSIS

		Matrix<T> ExpM1(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose elements will be used as exponents for Euler's constant

	DESCRIPTION

		This function raises Euler's constant to the power of the elements of a Matrix instance, and subtracts 1 from the result.

	RETURNS

		A Matrix instance containing the result of raising Euler's constant to the power of the elements of the Matrix instance,
		and subtracting 1 from the result.
	*/
	/**/
	template <typename T>
	Matrix<T> ExpM1(Matrix<T> a_matrix) {
		Matrix<T> temp = Exp(a_matrix);
		vector<vector<T>> result;
		for (int current_row = 0; current_row < temp.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < temp.GetCols(); current_column++) {
				// Here, we are subtracting 1 from the current value
				T current_value = temp.GetData()[current_row][current_column];
				row.push_back(current_value - 1);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ExpM1(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Log - Take the natural log of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Log(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance that will be used to take the natural log of its elements

	DESCRIPTION

		This function takes the natural log of the elements of a Matrix instance. To get an idea of how to take advantage of the C++ math library,
		the following link was consulted: https://www.geeksforgeeks.org/log-function-cpp/

	RETURNS

		A Matrix instance containing the result of taking the natural log of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Log(Matrix<T> a_matrix) {
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the natural log of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];

				// Make sure the current value is not less than or equal to 0
				if (current_value <= 0) {
					cout << "Error: Cannot take natural log of numbers less than or equal to 0." << endl;
					return Matrix<T>();
				}

				row.push_back(log(current_value));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Log(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Log10 - Take the base 10 logarithm of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Log10(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance that will be used to take the base 10 logarithm of its elements

	DESCRIPTION

		This function takes the base 10 logarithm of the elements of a Matrix instance. To get an idea of how to take advantage of the C++ math library,
		the following link was consulted: https://www.geeksforgeeks.org/log-function-cpp/

	RETURNS

		A Matrix instance containing the result of taking the base 10 logarithm of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Log10(Matrix<T> a_matrix) {
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the base 10 logarithm of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];

				// Make sure the current value is not less than or equal to 0
				if (current_value <= 0) {
					cout << "Error: Cannot take base 10 logarithm of numbers less than or equal to 0." << endl;
					return Matrix<T>();
				}

				row.push_back(log10(current_value));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Log10(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Log2 - Take the base 2 logarithm of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Log2(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance that will be used to take the base 2 logarithm of its elements

	DESCRIPTION

		This function takes the base 2 logarithm of the elements of a Matrix instance. To get an idea of how to take advantage of the C++ math library,
		the following link was consulted: https://www.geeksforgeeks.org/log2-function-in-c-with-examples/

	RETURNS

		A Matrix instance containing the result of taking the base 2 logarithm of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	Matrix<T> Log2(Matrix<T> a_matrix) {
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the base 2 logarithm of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];

				// Make sure the current value is not less than or equal to 0
				if (current_value <= 0) {
					cout << "Error: Cannot take base 2 logarithm of numbers less than or equal to 0." << endl;
					return Matrix<T>();
				}

				row.push_back(log2(current_value));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Log2(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		GCD - Calculate the greatest common divisor of the elements of a Matrix instance.

	SYNOPSIS

		int GCD(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose elements will be used to calculate the greatest common divisor

	DESCRIPTION

		This function computes the greatest common divisor of the elements of a Matrix instance.
		Help for implementing this function was found at the following link:
		https://www.geeksforgeeks.org/gcd-two-array-numbers/

	RETURNS

		The greatest common divisor of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	int GCD(Matrix<T> a_matrix) {
		vector<T> temp = Flatten(a_matrix);
		// To avoid a type conversion error, we cast the first element of the vector to type int
		int result = (int) temp[0];
		for (int current_element = 1; current_element < temp.size(); current_element++) {
			// Here, we call on the helper function responsible for finding the GCD of two numbers
			result = Find_GCD(result, (int) temp[current_element]);
		}
		return result;
	}
	/*int GCD(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		LCM - Calculate the least common multiple of the elements of a Matrix instance.

	SYNOPSIS

		int LCM(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance whose elements will be used to calculate the least common multiple

	DESCRIPTION

		This function computes the least common multiple of the elements of a Matrix instance.
		The formula for finding the least common multiple of two numbers 'a' and 'b' is: LCM(a, b) = (a * b) / GCD(a, b)
		Help for implementing this function was found at the following link:
		https://www.geeksforgeeks.org/lcm-of-given-array-elements/

	RETURNS

		The least common multiple of the elements of the Matrix instance.
	*/
	/**/
	template <typename T>
	int LCM(Matrix<T> a_matrix) {
		vector<T> temp = Flatten(a_matrix);
		// To avoid a type conversion error, we cast the first element of the vector to type int
		int result = (int) temp[0];
		for (int current_element = 1; current_element < temp.size(); current_element++) {
			// Here, we use the described formula for finding the LCM of two numbers
			result = (temp[current_element] * result) / Find_GCD(result, (int) temp[current_element]);
		}
		return result;
	}
	/*int LCM(Matrix<T> a_matrix);*/

	/*
	This is the helper function for finding the GCD of two numbers.
	Help for how to declare this function was found at the following links:
	https://stackoverflow.com/questions/5707324/using-non-template-functions-in-template-classes?rq=3
	https://www.geeksforgeeks.org/gcd-two-array-numbers/
	*/
	inline int Find_GCD(int a_num1, int a_num2) {
		if (a_num2 == 0) {
			// If we get here, then 'a_num1' represents the GCD
			return a_num1;
		}
		else if (a_num1 == 0) {
			// If we get here, then 'a_num2' represents the GCD
			return a_num2;
		}
		return Find_GCD(a_num2 % a_num1, a_num1);
	}

	/*
	This is the helper function for finding the factorial of a number.
	Help for how to declare this function was found at the following links:
	https://www.geeksforgeeks.org/program-for-factorial-of-a-number/
	*/
	inline int Factorial(int a_num) {
		if (a_num == 0 || a_num == 1) {
			return 1;
		}
		return a_num * Factorial(a_num - 1);
	}

	/*
	This is the helper function for finding the binomial coefficient of two numbers.
	Help for how to declare this function was found at the following links:
	https://www.geeksforgeeks.org/binomial-coefficient-dp-9/
	*/
	inline int Binomial_Coefficient(int a_num1, int a_num2) {
		if (a_num2 > a_num1) {
			return 0;
		}
		else if (a_num2 == 0 || a_num2 == a_num1) {
			return 1;
		}
		return Binomial_Coefficient(a_num1 - 1, a_num2 - 1) + Binomial_Coefficient(a_num1 - 1, a_num2);
	}

	/**/
	/*
	NAME

		Sin - Take the sine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Sin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the sine of its elements, all of which are in radians
			a_iterations -> The number of iterations to be used in the power series expansion of sine
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the sine of the elements of a Matrix instance. To do this, it uses the power series expansion of sine.
		Help for implementing this function was found at the following link:
		https://www.codingconnect.net/cpp-program-sine-series/
		And for help with rounding the result, a conversation with Microsoft Copilot was had.
		However, due to the limitations of the power series expansion of sine, and how this can lead to numbers
		that are too large for C++ to handle, this can lead to:
		- The result not being very precise.
		- The rounding mechanism not working as expected.

		This is why the user is allowed to specify the number of iterations to be used in the power series expansion of sine, and also why we usually
		utilize built-in functions for computing trigonometric functions.

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their sine values.
	*/
	/**/
	template <typename T>
	Matrix<T> Sin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places) {
		if (a_iterations <= 0 || a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of iterations and decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the sine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T sin_sum = 0;
				for (int current_iteration = 0; current_iteration < a_iterations; current_iteration++) {
					T next_term = (pow(-1, current_iteration) * pow(current_value, (2 * current_iteration) + 1)) / Factorial((2 * current_iteration) + 1);
					sin_sum += next_term;
				}

				// Round the result to the specified number of decimal places
				T rounded_value = round(sin_sum * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Sin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);*/

	/**/
	/*
	NAME

		Cos - Take the cosine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Cos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the cosine of its elements, all of which are in radians
			a_iterations -> The number of iterations to be used in the power series expansion of cosine
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the cosine of the elements of a Matrix instance, using the power series expansion of cosine.
		Help for implementing this function was found at the following link:
		https://www.codingconnect.net/cpp-program-cosine-series/
		However, due to the limitations of the power series expansion of cosine, and how this can lead to numbers
		that are too large for C++ to handle, this can lead to the same issues that were described for the 'Sin' function.
		Once again, this is why we usually utilize built-in functions for complex mathematical operations such as this.

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their cosine values.
	*/
	/**/
	template <typename T>
	Matrix<T> Cos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places) {
		if (a_iterations <= 0 || a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of iterations and decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the cosine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T cos_sum = 0;
				for (int current_iteration = 0; current_iteration < a_iterations; current_iteration++) {
					T next_term = (pow(-1, current_iteration) * pow(current_value, 2 * current_iteration)) / Factorial(2 * current_iteration);
					cos_sum += next_term;
				}

				// Round the result to the specified number of decimal places
				T rounded_value = round(cos_sum * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Cos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);*/

	/**/
	/*
	NAME

		Tan - Take the tangent of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Tan(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the tangent of its elements, all of which are in radians
			a_iterations -> The number of iterations to be used in the power series expansion of tangent
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the tangent of the elements of a Matrix instance. To do this,
		it takes advantage of the fact that tan(x) = sin(x) / cos(x) and calls on the 'Sin' and 'Cos' functions.

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their tangent values.
	*/
	/**/
	template <typename T>
	Matrix<T> Tan(Matrix<T> a_matrix, int a_iterations, int a_decimal_places) {
		if (a_iterations <= 0 || a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of iterations and decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}
		// Since tan(x) = sin(x) / cos(x), we can just divide the results of the 'Sin' and 'Cos' functions
		Matrix<T> sin_matrix = Sin(a_matrix, a_iterations, a_decimal_places), cos_matrix = Cos(a_matrix, a_iterations, a_decimal_places);
		Matrix<T> result = sin_matrix / cos_matrix;

		// Round the result to the specified number of decimal places
		vector<vector<T>> result_data = result.GetData();
		for (int current_row = 0; current_row < result.GetRows(); current_row++) {
			for (int current_column = 0; current_column < result.GetCols(); current_column++) {
				result_data[current_row][current_column] = round(result_data[current_row][current_column] * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
			}
		}

		result.SetData(result_data);
		return result;
	}
	/*Matrix<T> Tan(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcSin - Take the arcsine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcSin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the arcsine of its elements, all of which are in radians
			a_iterations -> The number of iterations to be used in the power series expansion of arcsine
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the arcsine of the elements of a Matrix instance. To do this, it uses the power series expansion of arcsine.
		Help for implementing the arcsine power series was found at the following link:
		https://math.stackexchange.com/questions/421892/how-to-calc-arc-sine-without-a-calculator
		Initially, there was some trouble involving the first part of the equation in the series because of how
		integer division works in C++, but ChatGPT helped me realize that I was supposed to be using floating point division instead.

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their arcsine values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcSin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places) {
		if (a_iterations <= 0 || a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of iterations and decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the arcsine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];

				// If the current value is outside of the range [-1, 1], then we cannot take its arcsine
				if (current_value > 1 || current_value < -1) {
					cout << "Error: Cannot take arcsine of value outside of range [-1, 1]." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				T arcsin_sum = 0;
				for (int current_iteration = 0; current_iteration < a_iterations; current_iteration++) {
					// To make the code more readable, we break up the formula for the power series expansion of arcsine into three parts
					T first_part = 2.0 / ((2 * current_iteration) + 1);
					int second_part = Binomial_Coefficient(2 * current_iteration, current_iteration);
					T third_part = pow(current_value / 2, 2 * current_iteration + 1);
					T next_term = first_part * second_part * third_part;
					arcsin_sum += next_term;
				}

				// Round the result to the specified number of decimal places
				T rounded_value = round(arcsin_sum * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcSin(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcCos - Take the arccosine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcCos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the arccosine of its elements, all of which are in radians
			a_iterations -> The number of iterations to be used in the power series expansion of arccosine
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the arccosine of the elements of a Matrix instance. To do this, it uses the power series expansion of arccosine.
		Help for implementing the arccosine power series was found at the following link:
		https://proofwiki.org/wiki/Power_Series_Expansion_for_Real_Arccosine_Function

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their arccosine values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcCos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places) {
		if (a_iterations <= 0 || a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of iterations and decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}
		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the arccosine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T pi_over_2 = (T) (g_pi / 2);

				// If the current value is outside of the range [-1, 1], then we cannot take its arccosine
				if (current_value > 1 || current_value < -1) {
					cout << "Error: Cannot take arccosine of value outside of range [-1, 1]." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				T arccos_sum = 0;
				for (int current_iteration = 0; current_iteration < a_iterations; current_iteration++) {
					// To make the code more readable, we break up the formula for the power series expansion of arccosine into three parts
					T first_part = Factorial(2.0 * current_iteration) / (pow(2, 2 * current_iteration) * pow(Factorial(current_iteration), 2));
					T second_part = pow(current_value, (2 * current_iteration) + 1) / ((2 * current_iteration) + 1);
					T next_term = first_part * second_part;
					arccos_sum += next_term;
				}

				T result = pi_over_2 - arccos_sum;
				// Round the result to the specified number of decimal places
				T rounded_value = round(result * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcCos(Matrix<T> a_matrix, int a_iterations, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcTan - Take the arctangent of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcTan(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the arctangent of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the arctangent of the elements of a Matrix instance. To do this, it uses equation #11
		described in the following publication: https://www-labs.iro.umontreal.ca/%7Emignotte/IFT2425/Documents/EfficientApproximationArctgFunction.pdf

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their arctangent values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcTan(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the arctangent of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				if (current_value > 1 || current_value < -1) {
					cout << "Error: Cannot take arctangent of value outside of range [-1, 1]." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}
			
				// Compute the arctangent, then round the result to the specified number of decimal places
				T arctanresult = current_value / (1 + (0.28086 * pow(current_value, 2)));
				T rounded_value = round(arctanresult * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcTan(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		Sinh - Take the hyperbolic sine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Sinh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic sine of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic sine of the elements of a Matrix instance. To do this, it uses the formula:
		sinh(x) = (1/2) * (e^x - e^-x). This formula was found at the following link:
		https://www.mathsisfun.com/sets/function-hyperbolic.html

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic sine values.
	*/
	/**/
	template <typename T>
	Matrix<T> Sinh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the hyperbolic sine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T sinh_value = (T)(0.5) * (pow(g_euler_constant, current_value) - pow(g_euler_constant, -current_value));
				T rounded_value = round(sinh_value * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Sinh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		Cosh - Take the hyperbolic cosine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Cosh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic cosine of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic cosine of the elements of a Matrix instance. To do this, it uses the formula:
		cosh(x) = (1/2) * (e^x + e^-x). This formula was found at the following link:
		https://www.mathsisfun.com/sets/function-hyperbolic.html

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic cosine values.
	*/
	/**/
	template <typename T>
	Matrix<T> Cosh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the hyperbolic cosine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T cosh_value = (T)(0.5) * (pow(g_euler_constant, current_value) + pow(g_euler_constant, -current_value));
				T rounded_value = round(cosh_value * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Cosh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		Tanh - Take the hyperbolic tangent of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> Tanh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic tangent of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic tangent of the elements of a Matrix instance. To do this, it uses the formula:
		tanh(x) = sinh(x) / cosh(x). This formula was found at the following link:
		https://numpy.org/doc/stable/reference/generated/numpy.tanh.html

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic tangent values.
	*/
	/**/
	template <typename T>
	Matrix<T> Tanh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		Matrix<T> sinh_matrix = Sinh(a_matrix, a_decimal_places), cosh_matrix = Cosh(a_matrix, a_decimal_places);
		Matrix<T> result = sinh_matrix / cosh_matrix;
		vector<vector<T>> result_data = result.GetData();
		for (int current_row = 0; current_row < result.GetRows(); current_row++) {
			for (int current_column = 0; current_column < result.GetCols(); current_column++) {
				// Here, we are rounding the hyperbolic tangent of the current value to the specified number of decimal places
				result_data[current_row][current_column] = round(result_data[current_row][current_column] * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
			}
		}
		result.SetData(result_data);
		return result;
	}
	/*Matrix<T> Tanh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcSinh - Take the hyperbolic arcsine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcSinh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic arcsine of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic arcsine of the elements of a Matrix instance. To do this, it uses the formula:
		arcsinh(x) = ln(x + sqrt((x^2) + 1)). This formula was found at the following link, which explains how NumPy's arcsinh() function
		can be used: https://btechgeeks.com/python-numpy-arcsinh-function/

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic arcsine values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcSinh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the hyperbolic arcsine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T ln_value = current_value + sqrt(pow(current_value, 2) + 1);

				// Make sure that the value is greater than 0, since we cannot take the arcsine of a value less than or equal to 0
				if (ln_value <= 0) {
					cout << "Error: Cannot take arcsine of value less than or equal to 0." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				// Round the result to the specified number of decimal places
				T arcsinh_value = log(ln_value);
				T rounded_value = round(arcsinh_value * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcSinh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcCosh - Take the hyperbolic arccosine of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcCosh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic arccosine of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic arccosine of the elements of a Matrix instance. To do this, it uses the formula:
		arccosh(x) = ln(x + sqrt((x^2) - 1)). This formula was found at the following link, which explains how NumPy's arccosh() function
		can be used: https://btechgeeks.com/python-numpy-arccosh-function/

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic arccosine values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcCosh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				// Here, we are taking the hyperbolic arccosine of the current value
				T current_value = a_matrix.GetData()[current_row][current_column];
				T ln_value = current_value + sqrt(pow(current_value, 2) - 1);

				// Make sure that the value is greater than 0, since we cannot take the arccosine of a value less than or equal to 0
				if (ln_value <= 0) {
					cout << "Error: Cannot take arccosine of value less than or equal to 0." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				// Round the result to the specified number of decimal places
				T arcsinh_value = log(ln_value);
				T rounded_value = round(arcsinh_value * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcCosh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		ArcTanh - Take the hyperbolic arctangent of the elements of a Matrix instance.

	SYNOPSIS

		Matrix<T> ArcTanh(Matrix<T> a_matrix, int a_decimal_places);
			a_matrix -> The Matrix instance that will be used to take the hyperbolic arctangent of its elements, all of which are in radians
			a_decimal_places -> The number of decimal places to be used in the result

	DESCRIPTION

		This function computes the hyperbolic arctangent of the elements of a Matrix instance. To do this, it uses the formula:
		arctanh(x) = (0.5) * ln((1 + x) / (1 - x)). This formula was found at the following link, which explains how NumPy's arctanh() function
		can be used: https://btechgeeks.com/python-numpy-arctanh-function/

	RETURNS

		A Matrix instance with all of the elements of the original Matrix instance replaced by their hyperbolic arctangent values.
	*/
	/**/
	template <typename T>
	Matrix<T> ArcTanh(Matrix<T> a_matrix, int a_decimal_places) {
		// Make sure that the number of decimal places is positive and non-zero
		if (a_decimal_places <= 0) {
			cout << "Error: Must have positive non-zero value for number of decimal places." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				T current_value = a_matrix.GetData()[current_row][current_column];

				// -1 and 1 are not included because the hyperbolic arctangents of these values are undefined
				if (current_value <= -1 || current_value >= 1) {
					cout << "Error: Cannot take hyperbolic arctangent of values greater than/equal to 1, or less than/equal to -1." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				// Make sure that the value to be used in the formula is positive
				T ln_value = (1 + current_value) / (1 - current_value);
				if (ln_value <= 0) {
					cout << "Error: Cannot use values less than/equal to 0 for ln(x)." << endl;
					cout << "Returning empty Matrix instance." << endl;
					return Matrix<T>();
				}

				T arctanh_value = (T)(0.5) * log((1 + current_value) / (1 - current_value));
				T rounded_value = round(arctanh_value * pow(10, a_decimal_places)) / pow(10, a_decimal_places);
				row.push_back(rounded_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> ArcTanh(Matrix<T> a_matrix, int a_decimal_places);*/

	/**/
	/*
	NAME

		Inner_Product - Compute the inner product of two matrices.

	SYNOPSIS

		Matrix<T> Inner_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used in the inner product
			a_matrix2 -> The second Matrix instance to be used in the inner product

	DESCRIPTION

		This function computes the inner product of two matrices.
		The inner product of two matrices is a new matrix with the number of rows equal to whichever matrix has more rows,
		and the number of columns equal to how many elements are in each row. Each element in the new matrix is the sum of the products
		of the corresponding elements in the rows of the two matrices.

	RETURNS

		A Matrix instance with the result of the inner product of the two input Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> Inner_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// When computing the inner product of two matrices, the number of columns must be the same for both matrices
		if (a_matrix1.GetCols() != a_matrix2.GetCols()) {
			cout << "Error: The number of columns in the first Matrix instance must be equal to the number of columns in the second Matrix instance." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		/*
			The result of the inner product of two matrices is a new matrix with the number of rows equal to whichever matrix has more rows,
			and the number of columns equal to how many elements are in each row.
		*/
		vector<vector<T>> result;
		for (int current_row_matrix_one = 0; current_row_matrix_one < a_matrix1.GetRows(); current_row_matrix_one++) {
			vector<T> result_row;
			vector<T> matrix_one_row = a_matrix1.GetData()[current_row_matrix_one];
			for (int current_row_matrix_two = 0; current_row_matrix_two < a_matrix2.GetRows(); current_row_matrix_two++) {
				vector<T> matrix_two_row = a_matrix2.GetData()[current_row_matrix_two];
				T inner_product = 0;
				for (int current_column = 0; current_column < a_matrix1.GetCols(); current_column++) {
					// The inner product of two vectors is the sum of the products of their corresponding elements
					inner_product += matrix_one_row[current_column] * matrix_two_row[current_column];
				}
				result_row.push_back(inner_product);
			}
			result.push_back(result_row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Inner_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Outer_Product - Compute the outer product of two matrices.

	SYNOPSIS

		Matrix<T> Outer_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used in the outer product
			a_matrix2 -> The second Matrix instance to be used in the outer product

	DESCRIPTION

		This function computes the outer product of two matrices. To do this, it iterates through the rows of the first matrix,
		and for each element in the row, it multiplies it by each element in the corresponding row of the second matrix.
		The result is a new matrix with the number of rows equal to the number of rows in the first matrix, and the number of columns
		equal to the number of columns in the second matrix.

	RETURNS

		A Matrix instance with the result of the outer product of the two input Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> Outer_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		vector<vector<T>> result;
		// We start by iterating through the rows of the first matrix
		for (int current_row_matrix_one = 0; current_row_matrix_one < a_matrix1.GetRows(); current_row_matrix_one++) {
			vector<T> matrix_one_row = a_matrix1.GetData()[current_row_matrix_one];
			for (int current_column_matrix_one = 0; current_column_matrix_one < a_matrix1.GetCols(); current_column_matrix_one++) {
				// Obtain the current value in the row of the first matrix
				T current_value_matrix_one = matrix_one_row[current_column_matrix_one];

				// Then, we iterate through the rows of the second matrix, multiplying the current value by each element in the second matrix's row
				for (int current_row_matrix_two = 0; current_row_matrix_two < a_matrix2.GetRows(); current_row_matrix_two++) {
					vector<T> result_row;
					vector<T> matrix_two_row = a_matrix2.GetData()[current_row_matrix_two];
					T outer_product = 0;
					for (int current_column_matrix_two = 0; current_column_matrix_two < a_matrix2.GetCols(); current_column_matrix_two++) {
						outer_product = current_value_matrix_one * matrix_two_row[current_column_matrix_two];
						result_row.push_back(outer_product);
					}
					result.push_back(result_row);
				}
			}
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Outer_Product(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Mat_Mul - Compute the matrix product.

	SYNOPSIS

		Matrix<T> Mat_Mul(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used in the matrix product
			a_matrix2 -> The second Matrix instance to be used in the matrix product

	DESCRIPTION

		This function computes the product of two matrices. To do this, it adds the products of the elements in the rows of the first matrix
		and the columns of the second matrix. The result is a new matrix with the number of rows equal to the number of rows in the first matrix,
		and the number of columns equal to the number of columns in the second matrix.

	RETURNS

		A Matrix instance with the result of the product of the two input Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> Mat_Mul(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// When multiplying two matrices, the number of columns in the first matrix must be equal to the number of rows in the second matrix
		if (a_matrix1.GetCols() != a_matrix2.GetRows()) {
			cout << "Error: The number of columns in the first Matrix instance must be equal to the number of rows in the second Matrix instance." << endl;
			cout << "Returning empty Matrix instance." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> result;
		for (int current_row_matrix_one = 0; current_row_matrix_one < a_matrix1.GetRows(); current_row_matrix_one++) {
			vector<T> result_row;
			vector<T> matrix_one_row = a_matrix1.GetData()[current_row_matrix_one];
			for (int current_column_matrix_two = 0; current_column_matrix_two < a_matrix2.GetCols(); current_column_matrix_two++) {
				// The product of two matrices is the sum of the products of the elements in the rows of the first matrix and the columns of the second matrix
				T product = 0;
				for (int current_column_matrix_one = 0; current_column_matrix_one < a_matrix1.GetCols(); current_column_matrix_one++) {
					product += matrix_one_row[current_column_matrix_one] * a_matrix2.GetData()[current_column_matrix_one][current_column_matrix_two];
				}
				result_row.push_back(product);
			}
			result.push_back(result_row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Mat_Mul(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Dot - Compute the dot product of two matrices.

	SYNOPSIS

		Matrix<T> Dot(Matrix<T> a_matrix1, Matrix<T> a_matrix2);
			a_matrix1 -> The first Matrix instance to be used in the dot product
			a_matrix2 -> The second Matrix instance to be used in the dot product

	DESCRIPTION

		This function computes the dot product of two matrices. If both matrices are 1D, then the result is the inner product of the two matrices.
		Otherwise, the result is the matrix product of the two matrices.

	RETURNS

		A Matrix instance with the result of the dot product of the two input Matrix instances.
	*/
	/**/
	template <typename T>
	Matrix<T> Dot(Matrix<T> a_matrix1, Matrix<T> a_matrix2) {
		// If we get here, then both matrices are 1D
		if ((a_matrix1.GetRows() == a_matrix2.GetRows()) && a_matrix1.GetRows() == 1) {
			return Inner_Product(a_matrix1, a_matrix2);
		}

		// If we get here, then both matrices are 2D
		else {
			return Mat_Mul(a_matrix1, a_matrix2);
		}
	}
	/*Matrix<T> Dot(Matrix<T> a_matrix1, Matrix<T> a_matrix2);*/

	/**/
	/*
	NAME

		Eigenvalue - Compute the eigenvalue of a matrix.

	SYNOPSIS

		pair<T, bool> Eigenvalue(Matrix<T> a_matrix, Matrix<T> a_eigenvector);
			a_matrix -> The Matrix instance for which the eigenvalue will be computed
			a_eigenvector -> The Matrix instance that serves as the eigenvector of the matrix

	DESCRIPTION

		This function computes the eigenvalue of a matrix. The equation for the eigenvalue Y is:
		AX = YX, where A is the matrix, X is the eigenvector, and Y is the eigenvalue. To solve for Y,
		we multiply the matrix by the eigenvector, then multiply the greatest common divisor of the product by the eigenvector.
		If the result of that is equal to the eigenvector times the matrix,
		then the eigenvalue is the greatest common divisor, and the equation is satisfied.
		However, this approach comes with some limitations:
			- The eigenvector is assumed to be a row vector, and was calculated correctly
			- The matrix is assumed to be square
			- The eigenvector is assumed to be the correct size
			- The eigenvalue is assumed to be positive

		To get an idea of how to work with pairs in C++, I used the following resources:
			- https://www.geeksforgeeks.org/pair-in-cpp-stl/
			- https://en.cppreference.com/w/cpp/utility/pair/make_pair

	RETURNS

		A pair with the first element being the eigenvalue and the second element being a boolean value indicating whether the eigenvalue satisfies the equation or not.
	*/
	/**/
	template <typename T>
	pair<T, bool> Eigenvalue(Matrix<T> a_matrix, Matrix<T> a_eigenvector) {
		// The matrix must be square
		if (a_matrix.GetCols() != a_matrix.GetRows()) {
			cout << "Error: Can only find eigenvalues of square matrices." << endl;
			return make_pair(0, false);
		}

		// The number of rows in the eigenvector must be equal to the number of columns in the matrix
		if (a_matrix.GetCols() != a_eigenvector.GetRows()) {
			cout << "Error: The number of rows in the eigenvector must be equal to the number of columns in the matrix." << endl;
			return make_pair(0, false);
		}

		// The eigenvector must be a row vector
		if (a_eigenvector.GetCols() > 1) {
			cout << "Error: The eigenvector must be a row vector." << endl;
			return make_pair(0, false);
		}

		// Multiply the matrix by the eigenvector, then obtain the greatest common divisor of the product
		Matrix<T> matrix_product = Mat_Mul(a_matrix, a_eigenvector);
		int greatest_common_divisor = GCD(matrix_product);
		vector<vector<T>> eigenvector_data = a_eigenvector.GetData(), matrix_product_data = matrix_product.GetData();

		for (int current_row = 0; current_row < eigenvector_data.size(); current_row++) {
			for (int current_column = 0; current_column < eigenvector_data[current_row].size(); current_column++) {
				// Multiply each value in the eigenvector by the greatest common divisor
				eigenvector_data[current_row][current_column] *= greatest_common_divisor;
			}
		}

		// Compare the new eigenvector to the product of the matrix and the original eigenvector
		if (eigenvector_data == matrix_product_data) {
			return make_pair(greatest_common_divisor, true);
		}
		else {
			return make_pair(greatest_common_divisor, false);
		}
	}
	/*pair<T, bool> Eigenvalue(Matrix<T> a_matrix, Matrix<T> a_eigenvector);*/

	/**/
	/*
	NAME

		Swap - Swap two elements in a 2D vector.

	SYNOPSIS

		void Swap(vector<vector<T>>& a_vector, int a_row_index_one, int a_column_index_one, int a_row_index_two, int a_column_index_two);
			a_vector -> The 2D vector in which the elements will be swapped
			a_row_index_one -> The row index of the first element
			a_column_index_one -> The column index of the first element
			a_row_index_two -> The row index of the second element
			a_column_index_two -> The column index of the second element

	DESCRIPTION

		This function swaps two elements in a 2D vector. It was adapted from the following source:
		https://github.com/mmarinaa/Determinant-Calculator/blob/master/src/DeterminantCalc.java

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Swap(vector<vector<T>>& a_vector, int a_row_index_one, int a_column_index_one, int a_row_index_two, int a_column_index_two) {
		// Make sure that the indices are within the boundaries of the 2D vector
		if ((a_row_index_one < 0 || a_row_index_one >= a_vector.size()) || (a_row_index_two < 0 || a_row_index_two >= a_vector.size()) ||
			(a_column_index_one < 0 || a_column_index_one >= a_vector[0].size()) || (a_column_index_two < 0 || a_column_index_two >= a_vector[0].size())) {
			cout << "Error: Cannot swap elements outside of the boundaries of the 2D vector." << endl;
			return;
		}

		T temp = a_vector[a_row_index_one][a_column_index_one];
		a_vector[a_row_index_one][a_column_index_one] = a_vector[a_row_index_two][a_column_index_two];
		a_vector[a_row_index_two][a_column_index_two] = temp;
	}
	/*void Swap(vector<vector<T>>& a_vector, int a_row_index_one, int a_column_index_one, int a_row_index_two, int a_column_index_two);*/

	/**/
	/*
	NAME

		Determinant - Compute the determinant of a matrix.

	SYNOPSIS

		T Determinant(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance for which the determinant will be computed

	DESCRIPTION

		This function computes the determinant of a matrix. It was adapted from the following source:
		https://github.com/mmarinaa/Determinant-Calculator/blob/master/src/DeterminantCalc.java
		Gemini (which is Google's version of ChatGPT) was consulted to help with commenting and understanding the code,
		since the comments in the original code were written in Ukrainian and the English translations
		did not help out that much.

	RETURNS

		The determinant of the matrix.
	*/
	/**/
	template <typename T>
	T Determinant(Matrix<T> a_matrix) {
		// The matrix must be square
		if (a_matrix.GetRows() != a_matrix.GetCols()) {
			cout << "Error: Can only compute the determinant of square matrices." << endl;
			return 0;
		}

		vector<vector<T>> matrix_data = a_matrix.GetData();
		T determinant = 1, number1, number2, total = 1;
		int index;
		vector<T> temp(a_matrix.GetRows() + 1);

		// Iterate through the rows of the matrix
		for (int i = 0; i < a_matrix.GetRows(); i++) {
			// Find the first non-zero element in the current row
			index = i;
			while (index < a_matrix.GetRows() && matrix_data[index][i] == 0) {
				index++;
			}

			// If the current row is all zeroes, then the determinant is 0
			if (index == a_matrix.GetRows()) {
				continue;
			}

			// If the first non-zero element is not in the right position, swap the rows
			if (index != i) {
				for (int j = 0; j < a_matrix.GetRows(); j++) {
					Swap(matrix_data, index, j, i, j);
				}
				determinant = determinant * pow(-1, index - i);
			}

			// Store the current row in a temporary vector
			for (int j = 0; j < a_matrix.GetRows(); j++) {
				temp[j] = matrix_data[i][j];
			}

			// Create zeros below the pivot element
			for (int j = i + 1; j < a_matrix.GetRows(); j++) {
				number1 = temp[i];
				number2 = matrix_data[j][i];
				for (int k = 0; k < a_matrix.GetRows(); k++) {
					// Subtract the multiple of the current row from the next row
					matrix_data[j][k] = (number1 * matrix_data[j][k]) - (number2 * temp[k]);
				}

				// Keep track of the product of the pivot elements
				total = total * number1;
			}
		}

		// Multiply the diagonal elements to get the determinant
		for (int i = 0; i < a_matrix.GetRows(); i++) {
			determinant = determinant * matrix_data[i][i];
		}

		// Divide the determinant by the product of the pivot elements, then return the result
		return (determinant / total);
	}
	/*T Determinant(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Rref - Compute the reduced row echelon form of a matrix.

	SYNOPSIS

		Matrix<T> Rref(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance for which the reduced row echelon form will be computed

	DESCRIPTION

		This function computes the reduced row echelon form of a matrix. It was adapted from the following source:
		https://rosettacode.org/wiki/Reduced_row_echelon_form#C#
		There were no comments in the original code, so GitHub Copilot was used to help with commenting and understanding the code.
	
	RETURNS

		A Matrix instance with the reduced row echelon form of the input matrix.
	*/
	/**/
	template <typename T>
	Matrix<T> Rref(Matrix<T> a_matrix) {
		vector<vector<T>> matrix_data = a_matrix.GetData();
		int lead = 0, row_count = matrix_data.size(), column_count = matrix_data[0].size();

		// Iterate through the rows of the matrix
		for (int current_row = 0; current_row < matrix_data.size(); current_row++) {
			// If the lead is greater than or equal to the number of columns, then we are done
			if (column_count <= lead) {
				break;
			}

			// Find the first non-zero element in the current row
			int i = current_row;
			while (matrix_data[i][lead] == 0) {
				i++;
				// If we reach the end of the rows, then we are done
				if (row_count == i) {
					// Reset the row index and move to the next column
					i = current_row;
					lead++;

					// If the lead is greater than or equal to the number of columns, then we are done
					if (column_count == lead) {
						lead--;
						break;
					}
				}
			}

			// Swap the current row with the row that has the first non-zero element
			for (int current_column = 0; current_column < column_count; current_column++) {
				T temp = matrix_data[current_row][current_column];
				matrix_data[current_row][current_column] = matrix_data[i][current_column];
				matrix_data[i][current_column] = temp;
			}

			// Divide the current row by the lead element
			T div = matrix_data[current_row][lead];
			if (div != 0) {
				for (int current_column = 0; current_column < column_count; current_column++) {
					matrix_data[current_row][current_column] /= div;
				}
			}

			// Subtract the multiple of the current row from the other rows
			for (int j = 0; j < row_count; j++) {
				if (j != current_row) {
					T sub = matrix_data[j][lead];
					for (int k = 0; k < column_count; k++) {
						matrix_data[j][k] -= (sub * matrix_data[current_row][k]);
					}
				}
			}

			// Move to the next column
			lead++;
		}
		return Matrix<T>(matrix_data);
	}
	/*Matrix<T> Rref(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Matrix_Power - Raise a matrix to the nth power.

	SYNOPSIS

		Matrix<T> Matrix_Power(Matrix<T> a_matrix, int a_power);
			a_matrix -> The Matrix instance that will be raised to the nth power
			a_power -> The power to which the matrix will be raised

	DESCRIPTION

		This function raises a matrix to the nth power. To do this, it recursively calls the 'Mat_Mul' function
		n - 1 times. Finally, it returns the original matrix if 'a_power' is
		set equal to a number less than or equal to 1. This approach was adapted from the following source:
		https://stackoverflow.com/questions/52490184/raising-a-matrix-to-the-nth-power

	RETURNS

		A Matrix instance with the result of raising the input matrix to the nth power.
	*/
	/**/
	template <typename T>
	Matrix<T> Matrix_Power(Matrix<T> a_matrix, int a_power) {
		// The matrix must be square
		if (a_matrix.GetCols() != a_matrix.GetRows()) {
			cout << "Error: Can only compute the power of square matrices." << endl;
			return Matrix<T>();
		}

		// If the power is less than or equal to 1, then return the original matrix
		if (a_power <= 1) {
			return a_matrix;
		}
		else {
			// Multiply the matrix by itself a_power - 1 times
			return Mat_Mul(Matrix_Power(a_matrix, a_power - 1), a_matrix);
		}
	}
	/*Matrix<T> Matrix_Power(Matrix<T> a_matrix, int a_power);*/

	/**/
	/*
	NAME

		Inverse - Compute the inverse of a matrix.

	SYNOPSIS

		Matrix<T> Inverse(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance for which the inverse will be computed

	DESCRIPTION

		This function computes the inverse of a matrix. To do this, it first checks if the matrix is square and if the determinant is non-zero.
		If both conditions are met, then it computes the inverse of the matrix like this:
			1. Create an identity matrix with the same number of rows as the input matrix
			2. Concatenate the input matrix and the identity matrix
			3. Compute the reduced row echelon form of the concatenated matrix
			4. Extract the right half of the concatenated matrix, which is the inverse of the input matrix

	RETURNS

		The inverse of the input matrix.
	*/
	/**/
	template <typename T>
	Matrix<T> Inverse(Matrix<T> a_matrix) {
		// The matrix must be square
		if (a_matrix.GetCols() != a_matrix.GetRows()) {
			cout << "Error: Can only compute the inverse of square matrices." << endl;
			return Matrix<T>();
		}

		// The determinant of the matrix must be non-zero
		if (Determinant(a_matrix) == 0) {
			cout << "Error: The determinant of the matrix is 0, so the matrix does not have an inverse." << endl;
			return Matrix<T>();
		}

		// Create an identity matrix with the same number of rows as the input matrix
		Matrix<T> identity_matrix = Identity<T>(a_matrix.GetRows());

		// Concatenate the input matrix and the identity matrix
		vector<Matrix<T>> matrices = { a_matrix, identity_matrix };
		Matrix<T> full_matrix = Concatenate(matrices, 1);

		// Compute the reduced row echelon form of the concatenated matrix
		Matrix<T> rref_matrix = Rref(full_matrix);
		vector<vector<T>> rref_data = rref_matrix.GetData(), result;

		// Extract the right half of the concatenated matrix, which is the inverse of the input matrix
		for (int current_row = 0; current_row < rref_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = a_matrix.GetCols(); current_column < rref_data[current_row].size(); current_column++) {
				T current_value = rref_data[current_row][current_column];
				row.push_back(current_value);
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Inverse(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		Median - Compute the median of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Median(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the median will be computed
			a_axis -> The axis along which the median will be computed

	DESCRIPTION

		This function computes the median of a matrix along a specified axis. If 'a_axis' is equal to 0, then the median is computed along the columns.
		If 'a_axis' is equal to 1, then the median is computed along the rows. Otherwise, the median is computed along the entire matrix.
		After sorting the elements along the specified axis, the median is computed using the following rules:
			- If the number of elements is odd, then the median is the middle element
			- If the number of elements is even, then the median is the average of the two middle elements
		The following link was used to help with understanding how to implement this function:
		https://www.geeksforgeeks.org/program-for-mean-and-median-of-an-unsorted-array/

	RETURNS

		A Matrix instance with the result of computing the median of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Median(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> result, matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// If we get here, then we are computing the median along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> column, result_row;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					column.push_back(matrix_data[current_row][current_column]);
				}
				sort(column.begin(), column.end());
				T median;
				if (column.size() % 2 == 0) {
					median = (column[(column.size() - 1) / 2] + column[column.size() / 2]) / 2;
				}
				else {
					median = column[column.size() / 2];
				}
				result_row.push_back(median);
				result.push_back(result_row);
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are computing the median along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row = matrix_data[current_row];
				sort(row.begin(), row.end());
				T median;
				if (row.size() % 2 == 0) {
					median = (row[(row.size() - 1) / 2] + row[row.size() / 2]) / 2;
				}
				else {
					median = row[row.size() / 2];
				}
				vector<T> result_row = { median };
				result.push_back(result_row);
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are computing the median along the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			sort(flattened_matrix.begin(), flattened_matrix.end());
			T median;
			if (flattened_matrix.size() % 2 == 0) {
				median = (flattened_matrix[(flattened_matrix.size() - 1) / 2] + flattened_matrix[flattened_matrix.size() / 2]) / 2;
			}
			else {
				median = flattened_matrix[flattened_matrix.size() / 2];
			}
			vector<T> result_row = { median };
			result.push_back(result_row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Median(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Mean - Compute the mean of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Mean(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the mean will be computed
			a_axis -> The axis along which the mean will be computed

	DESCRIPTION

		This function computes the mean of a matrix along a specified axis. If 'a_axis' is equal to 0, then the mean is computed along the columns.
		If 'a_axis' is equal to 1, then the mean is computed along the rows. Otherwise, the mean is computed along the entire matrix.
		After summing the elements along the specified axis, the mean is computed by dividing the sum by the number of elements.
		The following link was used to help with understanding how to implement this function:
		https://www.geeksforgeeks.org/program-for-mean-and-median-of-an-unsorted-array/

	RETURNS

		A Matrix instance with the result of computing the mean of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Mean(Matrix<T> a_matrix, int a_axis) {
		vector<vector<T>> result, matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// If we get here, then we are computing the mean along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> column, result_row;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					column.push_back(matrix_data[current_row][current_column]);
				}
				T sum = 0;
				for (int current_element = 0; current_element < column.size(); current_element++) {
					sum += column[current_element];
				}
				T mean = sum / column.size();
				result_row.push_back(mean);
				result.push_back(result_row);
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are computing the mean along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row = matrix_data[current_row];
				T sum = 0;
				for (int current_element = 0; current_element < row.size(); current_element++) {
					sum += row[current_element];
				}
				T mean = sum / row.size();
				vector<T> result_row = { mean };
				result.push_back(result_row);
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are computing the mean along the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			T sum = 0;
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				sum += flattened_matrix[current_element];
			}
			T mean = sum / flattened_matrix.size();
			vector<T> result_row = { mean };
			result.push_back(result_row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Mean(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Average - Compute the weighted average of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Average(Matrix<T> a_matrix, vector<T> a_weights, int a_axis);
			a_matrix -> The Matrix instance for which the weighted average will be computed
			a_weights -> The weights to be used in the weighted average
			a_axis -> The axis along which the weighted average will be computed

	DESCRIPTION

		This function computes the weighted average of a matrix along a specified axis. To do this, the following formula is used:
		weighted_average = sum(matrix * weights) / sum(weights)
		To help understand what this function is supposed to do, the following link was used:
		https://www.statology.org/numpy-average-vs-mean/

		NOTE: This implementation assumes that a 2D matrix with more than one row and more than one column is being used.

	RETURNS

		A Matrix instance with the result of computing the weighted average of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Average(Matrix<T> a_matrix, vector<T> a_weights, int a_axis) {
		T weight_sum = 0;

		// Sum the weights so we can divide by the result
		for (int current_weight_element = 0; current_weight_element < a_weights.size(); current_weight_element++) {
			weight_sum += a_weights[current_weight_element];
		}

		// Since we will be dividing by the sum of the weights, the sum of the weights must be greater than 0
		if (weight_sum == 0) {
			cout << "Error: The sum of the weights must be greater than 0." << endl;
			return Matrix<T>();
		}
		else {
			if (a_axis == 0) {
				// If we get here, then we are computing the average along the columns
				if (a_weights.size() != a_matrix.GetRows()) {
					cout << "Error: The number of weights must be equal to the number of rows in the matrix." << endl;
					return Matrix<T>();
				}

				vector<vector<T>> result;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					vector<T> row;
					T weighted_total = 0;
					for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
						// Multiply each element in the column by the corresponding weight, then sum the products
						T current_value = a_matrix.GetData()[current_row][current_column];
						weighted_total += current_value * a_weights[current_row];
					}

					// Divide the weighted total by the sum of the weights
					row.push_back(weighted_total / weight_sum);
					result.push_back(row);
				}
				return Matrix<T>(result);
			}
			else if (a_axis == 1) {
				// If we get here, then we are computing the average along the rows
				if (a_weights.size() != a_matrix.GetCols()) {
					cout << "Error: The number of weights must be equal to the number of columns in the matrix." << endl;
					return Matrix<T>();
				}

				vector<vector<T>> result;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					vector<T> row;
					T weighted_total = 0;
					for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
						// Multiply each element in the row by the corresponding weight, then sum the products
						T current_value = a_matrix.GetData()[current_row][current_column];
						weighted_total += current_value * a_weights[current_column];
					}

					// Divide the weighted total by the sum of the weights
					row.push_back(weighted_total / weight_sum);
					result.push_back(row);
				}
				return Matrix<T>(result);
			}
			else {
				// If we get here, then an invalid axis was specified
				cout << "Error: The axis must be equal to 0 or 1." << endl;
				return Matrix<T>();
			}
		}
	}
	/*Matrix<T> Average(Matrix<T> a_matrix, vector<T> a_weights, int a_axis);*/

	/**/
	/*
	NAME

		Std - Compute the standard deviation of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Std(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the standard deviation will be computed
			a_axis -> The axis along which the standard deviation will be computed

	DESCRIPTION

		This function computes the standard deviation of a matrix along a specified axis.
		The formula for standard deviation, as described on the link, https://numpy.org/doc/stable/reference/generated/numpy.std.html, is:
		std = sqrt(mean(abs(x - x.mean())**2))

	RETURNS

		A Matrix instance with the result of computing the standard deviation of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Std(Matrix<T> a_matrix, int a_axis) {
		if (a_axis == 0) {
			// If we get here, then we are computing the standard deviation along the columns
			Matrix<T> mean_matrix = Mean(a_matrix, 0);
			vector<vector<T>> result;
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> row;
				T sum = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					T current_value = a_matrix.GetData()[current_row][current_column];
					T mean = mean_matrix.GetData()[current_column][0];
					sum += pow(current_value - mean, 2);
				}
				row.push_back(sqrt(sum / a_matrix.GetRows()));
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are computing the standard deviation along the rows
			Matrix<T> mean_matrix = Mean(a_matrix, 1);
			vector<vector<T>> result;
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> row;
				T sum = 0;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					T current_value = a_matrix.GetData()[current_row][current_column];
					T mean = mean_matrix.GetData()[current_row][0];
					sum += pow(current_value - mean, 2);
				}
				row.push_back(sqrt(sum / a_matrix.GetCols()));
				result.push_back(row);
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are computing the standard deviation of the entire matrix
			Matrix<T> mean_matrix = Mean(a_matrix, -1);
			vector<vector<T>> result;
			vector<T> flattened_matrix = Flatten(a_matrix);
			T mean = mean_matrix.GetData()[0][0];
			T sum = 0;
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				sum += pow(flattened_matrix[current_element] - mean, 2);
			}
			vector<T> row = { sqrt(sum / flattened_matrix.size()) };
			result.push_back(row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> Std(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Var - Compute the variance of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Var(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the variance will be computed
			a_axis -> The axis along which the variance will be computed

	DESCRIPTION

		This function computes the variance of a matrix along a specified axis.
		Variance can be computed by simply raising the standard deviation to the power of 2.

	RETURNS

		A Matrix instance with the result of computing the variance of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Var(Matrix<T> a_matrix, int a_axis) {
		Matrix<T> sd_matrix;
		vector<vector<T>> result;

		// Obtain the appropriate standard deviation values depending on the axis
		if (a_axis == 0) {
			sd_matrix = Std(a_matrix, 0);
		}
		else if (a_axis == 1) {
			sd_matrix = Std(a_matrix, 1);
		}
		else {
			sd_matrix = Std(a_matrix, -1);
		}

		for (int current_row = 0; current_row < sd_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < sd_matrix.GetCols(); current_column++) {
				// Raise each value in 'sd_matrix' to the power of 2 to obtain variance
				T current_value = sd_matrix.GetData()[current_row][current_column];
				row.push_back(pow(current_value, 2));
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Var(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		PTP - Compute the peak-to-peak of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> PTP(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the peak-to-peak will be computed
			a_axis -> The axis along which the peak-to-peak will be computed

	DESCRIPTION

		This function computes the peak-to-peak of a matrix along a specified axis.
		The peak-to-peak is the difference between the maximum and minimum values along the specified axis.
		To help figure out how obtain the maximum and minimum values, the following links were consulted:
		https://en.cppreference.com/w/cpp/algorithm/max_element
		https://en.cppreference.com/w/cpp/algorithm/min_element

	RETURNS

		A Matrix instance with the result of computing the peak-to-peak of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> PTP(Matrix<T> a_matrix, int a_axis) {
		// 'temp' will be used for directly computing the peak-to-peak along the rows or columns
		vector<vector<T>> result, matrix_data = a_matrix.GetData();
		vector<T> temp;

		if (a_axis == 0) {
			// If we get here, then we are computing the peak-to-peak along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// 'temp' will contain the elements of the current column, so we can find the max and min
					temp.push_back(matrix_data[current_row][current_column]);
				}
				typename vector<T>::iterator max = max_element(temp.begin(), temp.end()), min = min_element(temp.begin(), temp.end());
				T max_element = *max, min_element = *min;
				vector<T> result_row = { max_element - min_element };
				result.push_back(result_row);
				temp.clear();
			}
			return Matrix<T>(result);
		}
		else if (a_axis == 1) {
			// If we get here, then we are computing the peak-to-peak along the columns
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					// 'temp' will contain the elements of the current row, so we can find the max and min
					temp.push_back(matrix_data[current_row][current_column]);
				}
				typename vector<T>::iterator max = max_element(temp.begin(), temp.end()), min = min_element(temp.begin(), temp.end());
				T max_element = *max, min_element = *min;
				vector<T> result_row = { max_element - min_element };
				result.push_back(result_row);
				temp.clear();
			}
			return Matrix<T>(result);
		}
		else {
			// If we get here, then we are computing the peak-to-peak along the entire matrix
			temp = Flatten(a_matrix);
			typename vector<T>::iterator max = max_element(temp.begin(), temp.end()), min = min_element(temp.begin(), temp.end());
			T max_element = *max, min_element = *min;
			vector<T> result_row = { max_element - min_element };
			result.push_back(result_row);
			return Matrix<T>(result);
		}
	}
	/*Matrix<T> PTP(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Percentile - Compute the percentile of a matrix along a specified axis.

	SYNOPSIS

		Matrix<T> Percentile(Matrix<T> a_matrix, double a_percentile, int a_axis);
			a_matrix -> The Matrix instance for which the percentile will be computed
			a_percentile -> The percentile to be computed
			a_axis -> The axis along which the percentile will be computed

	DESCRIPTION

		This function computes the percentile of a matrix along a specified axis.
		The formula for computing the percentile, 
		as described in a conversation with Microsoft Copilot is:
		percentile = (c - k) * x[f] + (k - f) * x[c]
		where: c = the index of the ceiling of the normalized ranking of the percentile
			   f = the index of the floor of the normalized ranking of the percentile
			   k = the normalized ranking of the percentile
			   x = the sorted elements of the matrix

		To help figure out how to implement this formula, the following link was consulted:
		https://code.activestate.com/recipes/511478-finding-the-percentile-of-the-values/

	RETURNS

		A Matrix instance with the result of computing the percentile of the input matrix along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<T> Percentile(Matrix<T> a_matrix, double a_percentile, int a_axis) {
		// Make sure that the percentile is between 0 and 1
		if (a_percentile < 0 || a_percentile > 1) {
			cout << "Error: The percentile must be between 0 and 1." << endl;
			return Matrix<T>();
		}

		vector<vector<T>> temp_data, matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// If we get here, then we are preparing to compute the percentile along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> temp;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// 'temp' will contain the elements of the current column, so we can find the percentile
					temp.push_back(matrix_data[current_row][current_column]);
				}
				sort(temp.begin(), temp.end());
				temp_data.push_back(temp);
			}
		}
		else if (a_axis == 1) {
			// If we get here, then we are preparing to compute the percentile along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> temp;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					// 'temp' will contain the elements of the current row, so we can find the percentile
					temp.push_back(matrix_data[current_row][current_column]);
				}
				sort(temp.begin(), temp.end());
				temp_data.push_back(temp);
			}
		}
		else {
			// If we get here, then we are preparing to compute the percentile along the entire matrix
			vector<T> temp = Flatten(a_matrix);
			sort(temp.begin(), temp.end());
			temp_data.push_back(temp);
		}

		// Compute the percentile using the formula explained in the function description
		vector<vector<T>> result;
		for (int current_vector = 0; current_vector < temp_data.size(); current_vector++) {
			double k = (temp_data[current_vector].size() - 1) * a_percentile;
			int f = floor(k), c = ceil(k);
			if (f == c) {
				vector<T> result_row = { temp_data[current_vector][int(k)] };
				result.push_back(result_row);
				continue;
			}
			T d0 = temp_data[current_vector][f] * (c - k), d1 = temp_data[current_vector][c] * (k - f);
			T result_value = d0 + d1;
			vector<T> result_row = { result_value };
			result.push_back(result_row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Percentile(Matrix<T> a_matrix, double a_percentile, int a_axis);*/

	/**/
	/*
	NAME

		Partition - Partition a vector.

	SYNOPSIS

		int Partition(vector<T>& a_vector, int a_start_index, int a_end_index);
			a_vector -> The vector to be partitioned
			a_start_index -> The start index of the vector
			a_end_index -> The end index of the vector

	DESCRIPTION

		This function partitions a vector. It was adapted from the following source:
		https://big-o.io/examples/quicksort/c++/

	RETURNS

		The index of the pivot element.
	*/
	/**/
	template <typename T>
	int Partition(vector<T>& a_vector, int a_start_index, int a_end_index) {
		int pivotindex = (a_start_index + a_end_index) / 2;
		T pivotvalue = a_vector[pivotindex];

		while (true) {
			// Find the first element that is greater than the pivot
			while (a_vector[a_start_index] < pivotvalue) {
				a_start_index++;
			}

			// Find the first element that is less than the pivot
			while (a_vector[a_end_index] > pivotvalue) {
				a_end_index--;
			}
			if (a_start_index >= a_end_index) {
				return a_end_index;
			}

			// Swap the elements at the start and end indices
			T temp = a_vector[a_start_index];
			a_vector[a_start_index] = a_vector[a_end_index];
			a_vector[a_end_index] = temp;
		}
	}
	/*int Partition(vector<T>& a_vector, int a_start_index, int a_end_index);*/

	/**/
	/*
	NAME

		Quick_Sort - Sort a vector using the quick sort algorithm.

	SYNOPSIS

		void Quick_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);
			a_vector -> The vector to be sorted
			a_start_index -> The start index of the vector
			a_end_index -> The end index of the vector

	DESCRIPTION

		This function performs the quick sort algorithm on a vector. It was adapted from the following source:
		https://big-o.io/examples/quicksort/c++/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Quick_Sort(vector<T>& a_vector, int a_start_index, int a_end_index) {
		// If the start index is less than the end index, then the vector has more than one element
		if (a_start_index < a_end_index) {
			// Compute the pivot index
			int pivot = Partition(a_vector, a_start_index, a_end_index);

			// Recursively call Quick_Sort on the left and right sides of the pivot index
			Quick_Sort(a_vector, a_start_index, pivot);
			Quick_Sort(a_vector, pivot + 1, a_end_index);
		}
	}
	/*void Quick_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);*/

	/**/
	/*
	NAME

		Heapify - Heapify a vector.

	SYNOPSIS

		void Heapify(vector<T>& a_vector, int a_vector_size, int a_index);
			a_vector -> The vector to be heapified
			a_vector_size -> The size of the vector
			a_index -> The index of the vector

	DESCRIPTION

		This function heapifies a vector. It was adapted from the following source:
		https://big-o.io/examples/heapsort/c++/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Heapify(vector<T>& a_vector, int a_vector_size, int a_index) {
		// Initialize the root, and its left and right children
		int max = a_index, left = 2 * a_index + 1, right = 2 * a_index + 2;

		// Set the maximum value to the left child if it is greater than the root
		if (left < a_vector_size && a_vector[left] > a_vector[max]) {
			max = left;
		}

		// Set the maximum value to the right child if it is greater than the root
		if (right < a_vector_size && a_vector[right] > a_vector[max]) {
			max = right;
		}

		// If the maximum value is not the root, then swap the root and the maximum value
		if (max != a_index) {
			T temp = a_vector[a_index];
			a_vector[a_index] = a_vector[max];
			a_vector[max] = temp;

			// Recursively call Heapify on the affected sub-tree
			Heapify(a_vector, a_vector_size, max);
		}
	}
	/*void Heapify(vector<T>& a_vector, int a_vector_size, int a_index);*/

	/**/
	/*
	NAME

		Heap_Sort - Sort a vector using the heap sort algorithm.

	SYNOPSIS

		void Heap_Sort(vector<T>& a_vector);
			a_vector -> The vector to be sorted

	DESCRIPTION

		This function performs the heap sort algorithm on a vector. It was adapted from the following source:
		https://big-o.io/examples/heapsort/c++/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Heap_Sort(vector<T>& a_vector) {
		int vector_size = a_vector.size();

		// Build the heap
		for (int current_element = vector_size / 2 - 1; current_element >= 0; current_element--) {
			Heapify(a_vector, vector_size, current_element);
		}

		// Extract elements from the heap one by one
		for (int current_element = vector_size - 1; current_element >= 0; current_element--) {

			// Move the current root to the end
			T temp = a_vector[0];
			a_vector[0] = a_vector[current_element];
			a_vector[current_element] = temp;

			// Call Heapify on the reduced heap
			Heapify(a_vector, current_element, 0);
		}
	}
	/*void Heap_Sort(vector<T>& a_vector);*/

	/**/
	/*
	NAME

		Merge - Merge two sub-vectors obtained from a larger vector.

	SYNOPSIS

		void Merge(vector<T>& a_vector, int a_start_index, int a_middle_index, int a_end_index);
			a_vector -> The vector that will be divided and merged
			a_start_index -> The start index of the vector
			a_middle_index -> The middle index of the vector
			a_end_index -> The end index of the vector

	DESCRIPTION

		This function divides a vector into two sub-vectors, then merges the two sub-vectors. It was adapted from the following source:
		https://big-o.io/examples/merge-sort/c++/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Merge(vector<T>& a_vector, int a_start_index, int a_middle_index, int a_end_index) {
		vector<T> left(a_middle_index - a_start_index + 1), right(a_end_index - a_middle_index);

		// Fill in the left and right vectors
		for (int current_element = 0; current_element < left.size(); current_element++) {
			left[current_element] = a_vector[a_start_index + current_element];
		}
		for (int current_element = 0; current_element < right.size(); current_element++) {
			right[current_element] = a_vector[a_middle_index + 1 + current_element];
		}

		int left_index = 0, right_index = 0, current_index = a_start_index;

		// Merge the left and right vectors
		while(left_index < left.size() && right_index < right.size()) {
			if (left[left_index] <= right[right_index]) {
				a_vector[current_index] = left[left_index];
				left_index++;
			}
			else {
				a_vector[current_index] = right[right_index];
				right_index++;
			}
			current_index++;
		}

		// Copy the remaining elements of the left vector
		while (left_index < left.size()) {
			a_vector[current_index] = left[left_index];
			left_index++;
			current_index++;
		}

		// Copy the remaining elements of the right vector
		while (right_index < right.size()) {
			a_vector[current_index] = right[right_index];
			right_index++;
			current_index++;
		}
	}
	/*void Merge(vector<T>& a_vector, int a_start_index, int a_middle_index, int a_end_index);*/

	/**/
	/*
	NAME

		Merge_Sort - Sort a vector using the merge sort algorithm.

	SYNOPSIS

		void Merge_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);
			a_vector -> The vector to be sorted
			a_start_index -> The start index of the vector
			a_end_index -> The end index of the vector

	DESCRIPTION

		This function performs the merge sort algorithm on a vector. It was adapted from the following source:
		https://big-o.io/examples/merge-sort/c++/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Merge_Sort(vector<T>& a_vector, int a_start_index, int a_end_index) {
		if (a_start_index < a_end_index) {
			// Find the middle index
			int middle_index = (a_start_index + a_end_index) / 2;

			// Recursively call Merge_Sort on the left and right halves
			Merge_Sort(a_vector, a_start_index, middle_index);
			Merge_Sort(a_vector, middle_index + 1, a_end_index);

			// Merge the two halves
			Merge(a_vector, a_start_index, middle_index, a_end_index);
		}
	}
	/*void Merge_Sort(vector<T>& a_vector, int a_start_index, int a_end_index);*/

	/**/
	/*
	NAME

		Insertion_Sort - Sort a vector using the insertion sort algorithm.

	SYNOPSIS

		void Insertion_Sort(vector<T>& a_vector, int a_left_index, int a_right_index);
			a_vector -> The vector to be sorted
			a_left_index -> The left index of the vector
			a_right_index -> The right index of the vector

	DESCRIPTION

		This function performs the insertion sort algorithm on a vector. It was adapted from the following source:
		https://www.geeksforgeeks.org/timsort/

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Insertion_Sort(vector<T>& a_vector, int a_left_index, int a_right_index) {
		for (int current_index = a_left_index + 1; current_index <= a_right_index; current_index++) {
			T temp = a_vector[current_index];
			int j = current_index - 1;

			// Move elements of a_vector[0... current_index - 1], that are greater than 'temp', to one position ahead of their current position
			while (j >= a_left_index && a_vector[j] > temp) {
				a_vector[j + 1] = a_vector[j];
				j--;
			}
			a_vector[j + 1] = temp;
		}
	}
	/*void Insertion_Sort(vector<T>& a_vector, int a_left_index, int a_right_index);*/

	/**/
	/*
	NAME

		Tim_Sort - Sort a vector using the Tim Sort algorithm.

	SYNOPSIS

		void Tim_Sort(vector<T>& a_vector, int a_run_size);
			a_vector -> The vector to be sorted
			a_run_size -> The run size

	DESCRIPTION

		This function performs the Tim Sort algorithm on a vector. The way that this algorithm works, as described at https://www.geeksforgeeks.org/timsort/
		(which is also where the code in the function was adapted from), is:
		1. Divide the array into blocks known as runs. We use insertion sort to sort the individual runs.
		2. Start with a run size of at least 32, and double the run size every time.
		3. Merge the runs using the 'Merge' function implemented earlier.
		4. Repeat the above steps until the entire array is sorted.

	RETURNS

		VOID
	*/
	/**/
	template <typename T>
	void Tim_Sort(vector<T>& a_vector, int a_run_size) {
		// The run size must be at least 32
		if (a_run_size < 32) {
			cout << "Error: The run size must be at least 32." << endl;
			return;
		}

		for (int current_index = 0; current_index < a_vector.size(); current_index += a_run_size) {
			// Create the runs using insertion sort
			Insertion_Sort(a_vector, current_index, min((current_index + a_run_size - 1), (int)a_vector.size() - 1));
		}

		// Start merging from 'a_run_size' size and double the size every time
		for (int current_size = a_run_size; current_size < a_vector.size(); current_size = 2 * current_size) {
			for (int left = 0; left < a_vector.size(); left += 2 * current_size) {
				int mid = left + current_size - 1;
				int right = min((left + 2 * current_size - 1), (int)a_vector.size() - 1);
				Merge(a_vector, left, mid, right);
			}
		}
	}
	/*void Tim_Sort(vector<T>& a_vector, int a_run_size);*/

	/**/
	/*
	NAME

		Sort - Sort a matrix using a specified algorithm along a specified axis.

	SYNOPSIS

		Matrix<T> Sort(Matrix<T> a_matrix, string a_kind, int a_axis);
			a_matrix -> The Matrix instance to be sorted
			a_kind -> The sorting algorithm to be used
			a_axis -> The axis along which the matrix will be sorted

	DESCRIPTION

		This function sorts a matrix using a specified algorithm along a specified axis. The sorting algorithms that can be used are:
			1. quicksort
			2. heapsort
			3. mergesort
			4. timsort

	RETURNS

		A sorted copy of the input matrix.
	*/
	/**/
	template <typename T>
	Matrix<T> Sort(Matrix<T> a_matrix, string a_kind, int a_axis) {
		vector<vector<T>> temp_data, matrix_data = a_matrix.GetData();

		// Prepare the data for sorting along the columns or rows
		if (a_axis == 0) {
			// If we get here, then we are sorting along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> temp_row;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					temp_row.push_back(matrix_data[current_row][current_column]);
				}
				temp_data.push_back(temp_row);
			}
		}
		else if (a_axis == 1) {
			// If we get here, then we are sorting along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> temp_row = matrix_data[current_row];
				temp_data.push_back(temp_row);
			}
		}
		else {
			// If we get here, then we are sorting the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			temp_data.push_back(flattened_matrix);
		}
	
		// Sort the data using the specified algorithm
		if (a_kind == "quicksort") {
			// If we get here, then we are using the quick sort algorithm
			for (int current_vector = 0; current_vector < temp_data.size(); current_vector++) {
				Quick_Sort(temp_data[current_vector], 0, temp_data[current_vector].size() - 1);
			}
		}
		else if (a_kind == "heapsort") {
			// If we get here, then we are using the heap sort algorithm
			for (int current_vector = 0; current_vector < temp_data.size(); current_vector++) {
				Heap_Sort(temp_data[current_vector]);
			}
		}
		else if (a_kind == "mergesort") {
			// If we get here, then we are using the merge sort algorithm
			for (int current_vector = 0; current_vector < temp_data.size(); current_vector++) {
				Merge_Sort(temp_data[current_vector], 0, temp_data[current_vector].size() - 1);
			}
		}
		else if (a_kind == "timsort") {
			// If we get here, then we are using the Tim sort algorithm
			for (int current_vector = 0; current_vector < temp_data.size(); current_vector++) {
				Tim_Sort(temp_data[current_vector], 32);
			}
		}
		else {
			cout << "Error: Invalid sorting algorithm." << endl;
			return Matrix<T>();
		}

		// Prepare the sorted data for returning
		vector<vector<T>> result;
		if (a_axis == 0) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<T> temp_row;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					temp_row.push_back(temp_data[current_row][current_column]);
				}
				result.push_back(temp_row);
			}
		}
		else if (a_axis == 1) {
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<T> temp_row = temp_data[current_row];
				result.push_back(temp_row);
			}
		}
		else {
			vector<T> flattened_matrix = temp_data[0];
			result.push_back(flattened_matrix);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Sort(Matrix<T> a_matrix, string a_kind, int a_axis);*/

	/**/
	/*
	NAME

		ArgMax - Find the index of the maximum element along a specified axis.

	SYNOPSIS

		Matrix<int> ArgMax(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the index of the maximum element will be found
			a_axis -> The axis along which the index of the maximum element will be found

	DESCRIPTION

		This function traverses the matrix along the specified axis to find the indices of the maximum elements.

	RETURNS

		A Matrix instance with the indices of the maximum elements along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<int> ArgMax(Matrix<T> a_matrix, int a_axis) {
		vector<vector<int>> result; 
		vector<vector<T>> matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// Here, we are finding the index of the maximum element along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<int> row;
				int max_index = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// If we get here, then we have not found the maximum element yet
					if (matrix_data[current_row][current_column] > matrix_data[max_index][current_column]) {
						max_index = current_row;
					}
				}
				row.push_back(max_index);
				result.push_back(row);
			}
		}
		else if (a_axis == 1) {
			// Here, we are finding the index of the maximum element along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<int> row;
				int max_index = 0;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (matrix_data[current_row][current_column] > matrix_data[current_row][max_index]) {
						// If we get here, then we have not found the maximum element yet
						max_index = current_column;
					}
				}
				row.push_back(max_index);
				result.push_back(row);
			}
		}
		else {
			// Here, we are finding the index of the maximum element in the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			int max_index = 0;
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				if (flattened_matrix[current_element] > flattened_matrix[max_index]) {
					// If we get here, then we have not found the maximum element yet
					max_index = current_element;
				}
			}
			vector<int> row = { max_index };
			result.push_back(row);
		}
		return Matrix<int>(result);
	}
	/*Matrix<int> ArgMax(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		ArgMin - Find the index of the minimum element along a specified axis.

	SYNOPSIS

		Matrix<int> ArgMin(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the index of the minimum element will be found
			a_axis -> The axis along which the index of the minimum element will be found

	DESCRIPTION

		This function traverses the matrix along the specified axis to find the indices of the minimum elements.

	RETURNS

		A Matrix instance with the indices of the minimum elements along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<int> ArgMin(Matrix<T> a_matrix, int a_axis) {
		vector<vector<int>> result;
		vector<vector<T>> matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// Here, we are finding the index of the minimum element along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<int> row;
				int min_index = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// If we get here, then we have not found the minimum element yet
					if (matrix_data[current_row][current_column] < matrix_data[min_index][current_column]) {
						min_index = current_row;
					}
				}
				row.push_back(min_index);
				result.push_back(row);
			}
		}
		else if (a_axis == 1) {
			// Here, we are finding the index of the minimum element along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<int> row;
				int min_index = 0;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (matrix_data[current_row][current_column] < matrix_data[current_row][min_index]) {
						// If we get here, then we have not found the minimum element yet
						min_index = current_column;
					}
				}
				row.push_back(min_index);
				result.push_back(row);
			}
		}
		else {
			// Here, we are finding the index of the minimum element in the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			int min_index = 0;
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				if (flattened_matrix[current_element] < flattened_matrix[min_index]) {
					// If we get here, then we have not found the minimum element yet
					min_index = current_element;
				}
			}
			vector<int> row = { min_index };
			result.push_back(row);
		}
		return Matrix<int>(result);
	}
	/*Matrix<int> ArgMin(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanArgMax - Find the index of the maximum element along a specified axis, ignoring nan values.

	SYNOPSIS

		Matrix<int> NanArgMax(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the index of the maximum element will be found
			a_axis -> The axis along which the index of the maximum element will be found

	DESCRIPTION

		This function traverses the matrix along the specified axis to find the indices of the maximum elements, ignoring nan values.
		To help figure out how to implement this function, a conversation with Microsoft Copilot was had.

	RETURNS

		A Matrix instance with the indices of the maximum elements along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<int> NanArgMax(Matrix<T> a_matrix, int a_axis) {
		vector<vector<int>> result;
		vector<vector<T>> matrix_data = a_matrix.GetData();

		if (a_axis == 0) {
			// Here, we are finding the index of the maximum element along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<int> row;
				int max_index = 0;
				T max_value = numeric_limits<T>::min();
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// If we get here, then we have not found the maximum element yet
					if (!isnan(matrix_data[current_row][current_column])) {
						if (matrix_data[current_row][current_column] > max_value) {
							max_value = matrix_data[current_row][current_column];
							max_index = current_row;
						}
					}
				}
				row.push_back(max_index);
				result.push_back(row);
			}
			return Matrix<int>(result);
		}
		else if (a_axis == 1) {
			// Here, we are finding the index of the maximum element along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<int> row;
				int max_index = 0;
				T max_value = numeric_limits<T>::min();
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (!isnan(matrix_data[current_row][current_column])) {
						if (matrix_data[current_row][current_column] > max_value) {
							// If we get here, then we have not found the maximum element yet
							max_value = matrix_data[current_row][current_column];
							max_index = current_column;
						}
					}
				}
				row.push_back(max_index);
				result.push_back(row);
			}
			return Matrix<int>(result);
		}
		else {
			// Here, we are finding the index of the maximum element in the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			int max_index = 0;
			T max_value = numeric_limits<T>::min();
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				if (!isnan(flattened_matrix[current_element])) {
					if (flattened_matrix[current_element] > max_value) {
						// If we get here, then we have not found the maximum element yet
						max_value = flattened_matrix[current_element];
						max_index = current_element;
					}
				}
			}
			vector<int> row = { max_index };
			result.push_back(row);
		}
		return Matrix<int>(result);
	}
	/*Matrix<int> NanArgMax(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NanArgMin - Find the index of the minimum element along a specified axis, ignoring nan values.

	SYNOPSIS

		Matrix<int> NanArgMin(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the index of the minimum element will be found
			a_axis -> The axis along which the index of the minimum element will be found

	DESCRIPTION

		This function traverses the matrix along the specified axis to find the indices of the minimum elements, ignoring nan values.
		To help figure out how to implement this function, a conversation with Microsoft Copilot was had.

	RETURNS

		A Matrix instance with the indices of the minimum elements along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<int> NanArgMin(Matrix<T> a_matrix, int a_axis) {
		vector<vector<int>> result;
		vector<vector<T>> matrix_data = a_matrix.GetData();

		if (a_axis == 0) {
			// Here, we are finding the index of the minimum element along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<int> row;
				int min_index = 0;
				T min_value = numeric_limits<T>::max();
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					// If we get here, then we have not found the minimum element yet
					if (!isnan(matrix_data[current_row][current_column])) {
						if (matrix_data[current_row][current_column] < min_value) {
							min_value = matrix_data[current_row][current_column];
							min_index = current_row;
						}
					}
				}
				row.push_back(min_index);
				result.push_back(row);
			}
			return Matrix<int>(result);
		}
		else if (a_axis == 1) {
			// Here, we are finding the index of the minimum element along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<int> row;
				int min_index = 0;
				T min_value = numeric_limits<T>::max();
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (!isnan(matrix_data[current_row][current_column])) {
						if (matrix_data[current_row][current_column] < min_value) {
							// If we get here, then we have not found the minimum element yet
							min_value = matrix_data[current_row][current_column];
							min_index = current_column;
						}
					}
				}
				row.push_back(min_index);
				result.push_back(row);
			}
			return Matrix<int>(result);
		}
		else {
			// Here, we are finding the index of the minimum element in the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			int min_index = 0;
			T min_value = numeric_limits<T>::max();
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				if (!isnan(flattened_matrix[current_element])) {
					if (flattened_matrix[current_element] < min_value) {
						// If we get here, then we have not found the minimum element yet
						min_value = flattened_matrix[current_element];
						min_index = current_element;
					}
				}
			}
			vector<int> row = { min_index };
			result.push_back(row);
		}
		return Matrix<int>(result);
	}
	/*Matrix<int> NanArgMin(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		NonZero - Find the indices of the non-zero elements in a matrix.

	SYNOPSIS

		tuple<vector<int>, vector<int>> NonZero(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance for which the indices of the non-zero elements will be found

	DESCRIPTION

		This function traverses the matrix to find the indices of the non-zero elements.
		It stores the row and column indices in two separate vectors, both of which
		are made part of a tuple that is then returned. To help gain an understanding of
		how to work with tuples, the following link was consulted:
		https://www.geeksforgeeks.org/tuples-in-c/

	RETURNS

		A tuple containing two vectors: the first vector contains the row indices of the non-zero elements,
		and the second vector contains the column indices of the non-zero elements.
	*/
	/**/
	template <typename T>
	tuple<vector<int>, vector<int>> NonZero(Matrix<T> a_matrix) {
		// Set up the vectors to hold the row and column indices, as well as the result tuple
		vector<int> row_indices, column_indices;
		tuple<vector<int>, vector<int>> result;

		for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				if (a_matrix.GetData()[current_row][current_column] != 0) {
					// If we get here, then the current element is non-zero
					row_indices.push_back(current_row);
					column_indices.push_back(current_column);
				}
			}
		}

		result = make_tuple(row_indices, column_indices);
		return result;
	}
	/*tuple<vector<int>, vector<int>> NonZero(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		FlatNonZero - Find the indices of the non-zero elements in a flattened matrix.

	SYNOPSIS

		Matrix<int> FlatNonZero(Matrix<T> a_matrix);
			a_matrix -> The Matrix instance for which the indices of the non-zero elements will be found after being flattened

	DESCRIPTION

		This function flattens the matrix and then traverses the flattened matrix to find the indices of the non-zero elements.

	RETURNS

		A Matrix instance containing the indices of the non-zero elements in the flattened input matrix.
	*/
	/**/
	template <typename T>
	Matrix<int> FlatNonZero(Matrix<T> a_matrix) {
		// Flatten the matrix and prepare the result vector
		vector<vector<int>> result;
		vector<T> flattened_matrix = Flatten(a_matrix);
		vector<int> temp;

		for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
			if (flattened_matrix[current_element] != 0) {
				// If we get here, then the current element is non-zero
				temp.push_back(current_element);
			}
		}
		result.push_back(temp);
		return Matrix<int>(result);
	}
	/*Matrix<int> FlatNonZero(Matrix<T> a_matrix);*/

	/**/
	/*
	NAME

		CountNonZero - Count the non-zero elements along a specified axis.

	SYNOPSIS

		Matrix<int> CountNonZero(Matrix<T> a_matrix, int a_axis);
			a_matrix -> The Matrix instance for which the non-zero elements will be counted
			a_axis -> The axis along which the non-zero elements will be counted

	DESCRIPTION

		This function counts the non-zero elements along the specified axis. If the axis is 0, then the non-zero elements
		along the columns are counted. If the axis is 1, then the non-zero elements along the rows are counted. If the axis
		is neither 0 nor 1, then the non-zero elements in the entire matrix are counted.

	RETURNS

		A Matrix instance containing the counts of the non-zero elements along the specified axis.
	*/
	/**/
	template <typename T>
	Matrix<int> CountNonZero(Matrix<T> a_matrix, int a_axis) {
		vector<vector<int>> result;
		vector<vector<T>> matrix_data = a_matrix.GetData();
		if (a_axis == 0) {
			// Here, we are counting the non-zero elements along the columns
			for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
				vector<int> row;
				int count = 0;
				for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
					if (matrix_data[current_row][current_column] != 0) {
						count++;
					}
				}
				row.push_back(count);
				result.push_back(row);
			}
		}
		else if (a_axis == 1) {
			// Here, we are counting the non-zero elements along the rows
			for (int current_row = 0; current_row < a_matrix.GetRows(); current_row++) {
				vector<int> row;
				int count = 0;
				for (int current_column = 0; current_column < a_matrix.GetCols(); current_column++) {
					if (matrix_data[current_row][current_column] != 0) {
						count++;
					}
				}
				row.push_back(count);
				result.push_back(row);
			}
		}
		else {
			// Here, we are counting the non-zero elements in the entire matrix
			vector<T> flattened_matrix = Flatten(a_matrix);
			vector<int> row;
			int count = 0;
			for (int current_element = 0; current_element < flattened_matrix.size(); current_element++) {
				if (flattened_matrix[current_element] != 0) {
					count++;
				}
			}
			row.push_back(count);
			result.push_back(row);
		}
		return Matrix<int>(result);
	}
	/*Matrix<int> CountNonZero(Matrix<T> a_matrix, int a_axis);*/

	/**/
	/*
	NAME

		Where - Return elements chosen from two matrices depending on a condition matrix.

	SYNOPSIS

		Matrix<T> Where(Matrix<bool> a_condition_matrix, Matrix<T> a_true, Matrix<T> a_false);
			a_condition_matrix -> The condition matrix
			a_true -> The matrix containing the values to choose from when the condition is true
			a_false -> The matrix containing the values to choose from when the condition is false

	DESCRIPTION

		This function returns elements chosen from two matrices depending on a condition matrix. If the current element in the
		condition matrix is true, then the corresponding element in the result matrix will be the current element in the true matrix.
		Otherwise, it will be the current element in the false matrix.

	RETURNS

		A Matrix instance containing the elements chosen from the true and false matrices depending on the condition matrix.
	*/
	/**/
	template <typename T>
	Matrix<T> Where(Matrix<bool> a_condition_matrix, Matrix<T> a_true, Matrix<T> a_false) {
		vector<vector<T>> result, true_data = a_true.GetData(), false_data = a_false.GetData();
		vector<vector<bool>> bool_data = a_condition_matrix.GetData();

		// Check if the condition matrix and the true and false matrices have the same dimensions
		if (a_condition_matrix.GetRows() != a_true.GetRows() || a_condition_matrix.GetCols() != a_true.GetCols()) {
			cout << "Error: The condition matrix and the true matrix must have the same dimensions." << endl;
			return Matrix<T>();
		}
		if (a_condition_matrix.GetRows() != a_false.GetRows() || a_condition_matrix.GetCols() != a_false.GetCols()) {
			cout << "Error: The condition matrix and the false matrix must have the same dimensions." << endl;
			return Matrix<T>();
		}

		// Prepare the result matrix
		for (int current_row = 0; current_row < a_condition_matrix.GetRows(); current_row++) {
			vector<T> row;
			for (int current_column = 0; current_column < a_condition_matrix.GetCols(); current_column++) {
				/*
					If the current element in the condition matrix is true, then the corresponding element in the result matrix
					will be the current element in the true matrix. Otherwise, it will be the current element in the false matrix.
				*/
				if (bool_data[current_row][current_column]) {
					row.push_back(true_data[current_row][current_column]);
				}
				else {
					row.push_back(false_data[current_row][current_column]);
				}
			}
			result.push_back(row);
		}
		return Matrix<T>(result);
	}
	/*Matrix<T> Where(Matrix<bool> a_condition_matrix, Matrix<T> a_true, Matrix<T> a_false);*/
}
#endif